/* Copyright (C) 2003, 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

   
#include "myadmin.h"
#include "MABackupPanel.h"
#include "MDataInterface.h"
#include "MDispatcher.h"
#include "MAdministrator.h"

#include "MGFileBrowserList.h"
#include "MGSchemaBrowserList.h"

#include "myg_utils.h"
#include "util.h"

#include <list>


#define PROFILE_SUFFIX ".mbp"


static bool backup_aborted;


static int backup_options[]={
  MYX_B_NO_CREATES,
  MYX_B_NO_EXTENDED_INSERT,
  MYX_B_ADD_DROP_TABLE,
  MYX_B_COMPLETE_INSERTS,
  MYX_B_COMMENT,
  MYX_B_DONT_WRITE_FULL_PATH,
  MYX_B_ANSI_QUOTES,
  MYX_B_DISABLE_KEYS,
  MYX_B_ADD_LOCKS//?
};

struct ProgressData {
  Glib::ustring table;
  int num_tables_total;
  int num_tables;
  int num_rows_total;
  int num_rows;
  bool ready;
};


struct BackupThreadArgument {
  MABackupPanel *me;
  std::string path;
  MYX_BACKUP_PROFILE *profile;
  ProgressData pdata;
};




MABackupPanel::MABackupPanel(MAdministrator *app, MDataInterface *data)
    : MAPanel(app, data), _dirty(false)
{
}


MABackupPanel::~MABackupPanel()
{
  delete _progress_xml;	
}



void MABackupPanel::setup_selection_list()
{
  Gtk::TreeView *tree= get_tree("selection_tree");
  Gtk::TreeView::Column *column;
  MGCellRendererTristate::ImageList image_list(3);
  
  Gdk::Color color;
  // load tristate button image list
  image_list[0]= Gdk::Pixmap::create_from_xpm(_app->window()->get_window(),
                                              color,
                                              APP_DATADIR"/option_unchecked.xpm");
  image_list[1]= Gdk::Pixmap::create_from_xpm(_app->window()->get_window(),
                                              color,
                                              APP_DATADIR"/option_checked_light.xpm");
  image_list[2]= Gdk::Pixmap::create_from_xpm(_app->window()->get_window(),
                                              color,
                                              APP_DATADIR"/option_checked.xpm");
  
  // setup list columns

  _selection_tree= tree;
  
  _selection_store= Gtk::TreeStore::create(_sel_columns);
  tree->set_model(_selection_store);

  MGCellRendererTristate *trend= new MGCellRendererTristate(image_list);
  column= new Gtk::TreeView::Column(_("Objects"), *Gtk::manage(trend));
  tree->append_column(*Gtk::manage(column));
  column->add_attribute(trend->property_state(), _sel_columns._state);
  trend->signal_toggled().connect(SigC::slot(*this,&MABackupPanel::row_toggled));
  
  column= tree->get_column(0);
  column->pack_start(_sel_columns._icon, false);
  column->pack_start(_sel_columns._object);
  

  tree->append_column(_("Table Type"), _sel_columns._type);
  tree->append_column(_("Rows"), _sel_columns._rows);
  tree->append_column(_("Data Length"), _sel_columns._length);
  tree->append_column(_("Update Time"), _sel_columns._time);
  
  tree->get_column(1)->set_alignment(1.0);
  tree->get_column(2)->set_alignment(1.0);

  Glib::RefPtr<Gtk::TreeSelection> sel= tree->get_selection();

  sel->signal_changed().connect(SigC::slot(*this,&MABackupPanel::selected_object));
  
}


bool MABackupPanel::init()
{
  std::string profile_dir= prefs.build_path_to(prefs.backup_profiles_directory);

  if (_xml)
    return true;

  if (!MAPanel::init_from_glade(GLADE_DATADIR"/"GLADE_BACKUP_FILE, "panel_frame"))
    return false;

  _progress_xml= new MGGladeXML(GLADE_DATADIR"/"GLADE_BACKUP_FILE, "progress_dialog");
  if (!_progress_xml)
  {
    g_error(_("Could not load backup progress dialog"));
    return false;
  }

  _progress_label= (Gtk::Label*)_progress_xml->get_widget("progress_label");
  _table_progress= (Gtk::ProgressBar*)_progress_xml->get_widget("table_progress");
  _row_progress= (Gtk::ProgressBar*)_progress_xml->get_widget("row_progress");

  _file_browser= new MGFileBrowserList(_("Backup Profiles"),
                                       profile_dir);
  _file_browser->set_extension(PROFILE_SUFFIX);
  _file_browser->signal_selected().connect(SigC::slot(*this,&MABackupPanel::profile_selected));

  {
    Gtk::Menu *menu= Gtk::manage(new Gtk::Menu);
    Gtk::MenuItem *item;
    Gtk::Image *image;

    image= Gtk::manage(new Gtk::Image(Gtk::Stock::NEW, Gtk::ICON_SIZE_MENU));
    item= Gtk::manage(new Gtk::ImageMenuItem(*image, _("_New Profile"), true));
    item->signal_activate().connect(SigC::slot(*this,&MABackupPanel::create_profile));
    image->show();
    item->show();
    menu->append(*item);

    image= Gtk::manage(new Gtk::Image(Gtk::Stock::COPY, Gtk::ICON_SIZE_MENU));
    item= Gtk::manage(new Gtk::ImageMenuItem(*image, _("_Clone Profile"), true));
    item->signal_activate().connect(SigC::slot(*this,&MABackupPanel::clone_profile));
    image->show();
    item->show();
    menu->append(*item);

    image= Gtk::manage(new Gtk::Image(Gtk::Stock::DELETE, Gtk::ICON_SIZE_MENU));
    item= Gtk::manage(new Gtk::ImageMenuItem(*image, _("_Delete Profile"), true));
    item->signal_activate().connect(SigC::slot(*this,&MABackupPanel::delete_profile));
    image->show();
    item->show();
    menu->append(*item);
    
    _file_browser->set_popup_menu(menu);
  }
  
  setup_selection_list();
  
  _schema_icon= _app->load_pixbuf("16x16_Database.png");
  _sys_schema_icon= _schema_icon;

  _table_icon= _app->load_pixbuf("16x16_Table.png");
  _column_icon= _app->load_pixbuf("16x16_Field.png");

  _schema_browser= new MGSchemaBrowserList("", true);
  ((Gtk::Container*)get_widget("schemata_frame"))->add(*Gtk::manage(_schema_browser));
  _schema_browser->signal_selected().connect(SigC::slot(*this,&MABackupPanel::schema_selected));
  _schema_browser->set_icons(_schema_icon, _sys_schema_icon, _table_icon, _column_icon);
  _schema_browser->show();

  get_button("target_button")->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::select_target_path));

  // setup signals
  Gtk::Button *btn;
  
  btn= (Gtk::Button*)_progress_xml->get_widget("stop_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::stop_backup));
  
  btn= get_button("add_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::add_object));

  btn= get_button("remove_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::remove_object));

  btn= get_button("new_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::create_profile));

  btn= get_button("import_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::import_profile));

  btn= get_button("save_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::save_profile));

  btn= get_button("start_button");
  btn->signal_clicked().connect(SigC::slot(*this,&MABackupPanel::perform_backup));

  get_entry("name_entry")->signal_changed().connect(SigC::slot(*this,&MABackupPanel::set_dirty));
  get_entry("name_entry")->signal_changed().connect(SigC::slot(*this,&MABackupPanel::name_changed));
  
  for (int i= 1; i <= 8; i++)
  {
    char buf[32];
    sprintf(buf, "option%i_check", i);
    ((Gtk::CheckButton*)get_widget(buf))->signal_toggled().connect(SigC::slot(*this,&MABackupPanel::set_dirty));
  }

  ((Gtk::RadioButton*)get_widget("lockall_radio"))->signal_toggled().connect(SigC::slot(*this,&MABackupPanel::set_dirty));
  ((Gtk::RadioButton*)get_widget("singletrans_radio"))->signal_toggled().connect(SigC::slot(*this,&MABackupPanel::set_dirty));
  ((Gtk::RadioButton*)get_widget("normal_radio"))->signal_toggled().connect(SigC::slot(*this,&MABackupPanel::set_dirty));

  get_toggle("backup_all_check")->signal_toggled().connect(SigC::slot(*this,&MABackupPanel::set_dirty));
  
  ((Gtk::OptionMenu*)get_widget("schedule_menu"))->signal_changed().connect(SigC::slot(*this,&MABackupPanel::schedule_type_changed));
  schedule_type_changed();
  
  update_sensitivity(false);

  return true;
}


void MABackupPanel::show()
{
  MAPanel::show();
  
  _schema_browser->set_catalogs(_data->get_catalogs());
  _file_browser->refresh();
}


bool MABackupPanel::before_show()
{
  _app->add_side_panel(_file_browser);

  return true;
}


bool MABackupPanel::before_hide()
{  
  _app->remove_side_panel(_file_browser);

  return true;
}


/**********************************************************************/

void MABackupPanel::name_changed()
{
  Glib::ustring name= get_entry("name_entry")->get_text();
  Glib::ustring oprefix= get_entry("prefix_entry")->get_text();

  //XXX
  if (oprefix == name.substr(0, name.size()-1))
  {
//XXX    get_entry("prefix_entry")->set_text(name);
  }
}


void MABackupPanel::update_sensitivity(bool flag)
{
  get_widget("notebook")->set_sensitive(flag);
  get_widget("save_button")->set_sensitive(flag);
  
  update_button_bar_sensitivity();
}

void MABackupPanel::update_button_sensitivity()
{
  Glib::RefPtr<Gtk::TreeSelection> sel= _selection_tree->get_selection();
  Gtk::TreeIter iter;
  Gtk::TreeModel::Row row;

  if (sel->count_selected_rows())
  {
    get_widget("add_button")->set_sensitive(false);

    iter= sel->get_selected();
    
    bool flag= false;
    if (iter)
    {
      row= *iter;
      flag= row[_sel_columns._what] == SelectionColumns::CSchema;
    }
    // we can only remove whole schemas
    get_widget("remove_button")->set_sensitive(flag);
  }
  else
  {
    get_widget("remove_button")->set_sensitive(false);

    Glib::ustring catalog, schema, foo;
    _schema_browser->get_selected_schema(iter, catalog, schema, foo);
    
    // check if the selected schema is already in the list
    Gtk::TreeModel::Children ch= _selection_store->children();
    bool found= false;
    
    for (iter= ch.begin(); iter != ch.end(); ++iter)
    {
      row= *iter;
      
      if (schema == row[_sel_columns._object])
      {
        found= true;
        break;
      }
    }
    // if schema not yet in list, allow it to be added
    get_widget("add_button")->set_sensitive(!found);
  }
}


void MABackupPanel::select_target_path()
{
  Gtk::FileSelection fsel(_("Select Target Directory"));

  fsel.hide_fileop_buttons();
  fsel.set_filename(get_entry("target_entry")->get_text()+"/");
  fsel.get_file_list()->set_sensitive(false);

  if (fsel.run() == Gtk::RESPONSE_OK)
  {
    get_entry("target_entry")->set_text(fsel.get_filename());
  }
}


void MABackupPanel::update_button_bar_sensitivity()
{
  bool ok= false;

  if (get_widget("notebook")->is_sensitive())
  {
    if (!get_toggle("backup_all_check")->get_active())
    {
      // check whether there's some selected table
      Gtk::TreeModel::Children ch= _selection_store->children();
      for (Gtk::TreeIter p= ch.begin(); !ok && p != ch.end(); ++p)
      {
        Gtk::TreeModel::Row pr= *p;
        Gtk::TreeModel::Children cch= pr.children();
        for (Gtk::TreeIter i= cch.begin(); i != cch.end(); ++i)
        {
          Gtk::TreeModel::Row cr= *i;
          if (cr[_sel_columns._state] == MGCellRendererTristate::ON)
          {
            ok= true;
            break;
          }
          break;
        }
      }
    }
    else
      ok= true;
  }

  get_widget("start_button")->set_sensitive(ok);
}


void MABackupPanel::add_schema_to_list(const Glib::ustring &catalog,
                                       const Glib::ustring &schema,
                                       bool checked)
{ 
  Gtk::TreeIter s_iter, t_iter;
  Gtk::TreeModel::Row s_row, t_row;
  unsigned long total_rows= 0, total_length= 0;

  // add schema
  s_iter= _selection_store->append();
  s_row= *s_iter;

  s_row[_sel_columns._icon]= _schema_icon;
  s_row[_sel_columns._object]= schema;
  s_row[_sel_columns._state]= checked?MGCellRendererTristate::ON:MGCellRendererTristate::OFF;
  s_row[_sel_columns._what]= SelectionColumns::CSchema;

  // fetch table list for schema
  MYX_SCHEMA_TABLE_STATUS *tables= _data->get_schema_table_status(catalog,schema);

  // add tables of this schema
  for (unsigned int i= 0; tables && i < tables->schema_tables_num; ++i)
  {
    MYX_TABLE_STATUS *table= &tables->schema_tables[i];
    t_iter= _selection_store->append(s_row.children());
    t_row= *t_iter;

    t_row[_sel_columns._icon]= _table_icon;
    t_row[_sel_columns._object]= (char*)table->table_name;
    t_row[_sel_columns._type]= (char*)table->table_type;
    t_row[_sel_columns._rows]= atol((char*)table->rows);
    t_row[_sel_columns._length]= format_value(atol((char*)table->data_length));
    t_row[_sel_columns._time]= (char*)table->update_time;

    t_row[_sel_columns._lengthn]= atol((char*)table->data_length);
    t_row[_sel_columns._state]= checked?MGCellRendererTristate::ON:MGCellRendererTristate::OFF;
    t_row[_sel_columns._what]= SelectionColumns::CTable;
    
    total_rows+= atol((char*)table->rows);
    total_length+= atol((char*)table->data_length);
  }

  s_row[_sel_columns._rows]= total_rows;
  s_row[_sel_columns._lengthn]= total_length;
  s_row[_sel_columns._length]= format_value(total_length);
  
  myx_free_schema_table_status(tables);
}


void MABackupPanel::add_object()
{
  Gtk::TreeIter node;
  Glib::ustring catalog;
  Glib::ustring schema;
  Glib::ustring dummy;

  if (_schema_browser->get_selected_schema(node, catalog, schema, dummy))
  {
    add_schema_to_list(catalog, schema, true);
  }

  update_button_bar_sensitivity();
  set_dirty();
}


void MABackupPanel::remove_object()
{
  Gtk::TreeIter iter= _selection_tree->get_selection()->get_selected();
  
  _selection_store->erase(iter);
  
  update_button_bar_sensitivity();
  set_dirty();
}


void MABackupPanel::selected_object()
{
  _schema_browser->unselect();

  update_button_sensitivity();
}


void MABackupPanel::copy_profile(const std::string &source_name,
                                 const std::string &dest_name)
{
  std::string profile_dir= _file_browser->get_directory()+"/";
  std::string new_name;
  MYX_BACKUP_PROFILE *profile;
  MYX_ADMIN_LIB_ERROR err;
  int i;

  // copy the contents to the profile directory
  profile= myx_load_profile(Glib::path_get_basename(source_name).c_str(), 
                            std::string(Glib::path_get_dirname(source_name)+"/").c_str(),
                            &err);
  if (!profile)
  {
    show_adminlib_error(*_app->window(), 
                        ufmt(_("Could not read the profile file '%s'"),
                             source_name.c_str()),
                        err);
    return;
  }
  i= 1;

  new_name= dest_name;
  while (_file_browser->check_exists(new_name))
    new_name= dest_name+tostr(i++);

  g_free(profile->profile_name);
  profile->profile_name= g_strdup(new_name.c_str());

  new_name+= PROFILE_SUFFIX;
  
  if (myx_save_profile(new_name.c_str(),
                       profile_dir.c_str(),
                       profile) < 0)
  {
    myx_free_profile(profile);
    show_error(*_app->window(), 
               _("Could not write the copied profile file"));
    return;
  }
  myx_free_profile(profile);
}


void MABackupPanel::import_profile()
{
  Gtk::FileSelection fsel;

  if (fsel.run() == Gtk::RESPONSE_OK)
  {
    std::string name= Glib::path_get_basename(fsel.get_filename());
    if (name.rfind('.')!=std::string::npos)
    {
      name= name.substr(0, name.rfind('.'));
    }

    copy_profile(fsel.get_filename(), name);

    _file_browser->refresh();
  }
}


void MABackupPanel::create_profile()
{
  std::string name= "profile";

  int i= 1;
  // figure a new name
  while (_file_browser->check_exists(name))
  {
    name= "profile_"+tostr(i++);
  }

  _file_browser->unselect();

  get_entry("name_entry")->set_text(name);
  get_entry("prefix_entry")->set_text(name);

  // add to the list
  save_profile();

  Gtk::TreeIter node;
  _file_browser->refresh();
  if (_file_browser->find_node(name, _file_browser->_columns._name, node))
    _file_browser->set_selection(node);
}


void MABackupPanel::clone_profile()
{
  Gtk::TreeIter iter= _file_browser->get_selected();
  std::string profile_dir= _file_browser->get_directory()+"/";
  std::string name, new_name;

  _file_browser->get_row_object(iter, name);

  copy_profile(_file_browser->get_path_for_entry(name),
               name+"_copy");

  _file_browser->refresh();
}


void MABackupPanel::delete_profile()
{
  Gtk::TreeIter iter= _file_browser->get_selected();
  std::string name;

  _file_browser->get_row_object(iter, name);

  _file_browser->remove(name);
  _file_browser->refresh();
}


void MABackupPanel::show_profile(MYX_BACKUP_PROFILE *profile)
{
  if (profile)
  {
    _profile_name= profile->profile_name;
  
    get_entry("name_entry")->set_text(profile->profile_name?:"");
    get_entry("target_entry")->set_text(profile->target_directory?:"");
    get_entry("prefix_entry")->set_text(profile->filename_prefix?:"");
  }
  else
  {
    _profile_name.clear();
    get_entry("name_entry")->set_text("");
    get_entry("target_entry")->set_text("");
    get_entry("prefix_entry")->set_text("");
  }
  for (int i= 1; i <= 8; i++)
  {
    char buf[32];
    Gtk::CheckButton *check;
    sprintf(buf, "option%i_check", i);
    check= (Gtk::CheckButton*)get_widget(buf);

    check->set_active(profile && (profile->options & backup_options[i-1])!=0);
  }
  
  get_toggle("backup_all_check")->set_active(profile && (profile->options & MYX_B_COMPLETE_SCHEMATAS));
  if (profile && profile->options & MYX_B_SINGLE_TRANSACTION)
    get_toggle("singletrans_radio")->set_active(true);
  else if (profile && profile->options & MYX_B_LOCK_ALL_TABLES)
    get_toggle("lockall_radio")->set_active(true);
  else
    get_toggle("normal_radio")->set_active(true);

  _selection_store->clear();

  
  if (profile)
  {
    std::map<Glib::ustring,unsigned int> seen;

    for (unsigned int i= 0; i < profile->backup_content->tables_num; i++)
    {
      MYX_BACKUP_TABLE *table= profile->backup_content->tables+i;
      Glib::ustring key(Glib::ustring(table->catalog)+"."+Glib::ustring(table->schema));
      
      if (seen.find(key) == seen.end())
      {
        seen[key]= 1;
        
        add_schema_to_list(table->catalog, table->schema, false);
      }
      else
      {
        // count how many tables are selected
        seen[key]++;
      }
    }

    Gtk::TreeModel::Children top_nodes= _selection_store->children();

    for (unsigned int i= 0; i < profile->backup_content->tables_num; i++)
    {
      MYX_BACKUP_TABLE *table= profile->backup_content->tables+i;
      Gtk::TreeIter piter, citer;
      Gtk::TreeModel::Row prow, crow;
      Glib::ustring catalog(table->catalog?:"def");
      Glib::ustring schema(table->schema);
      Glib::ustring key(catalog+"."+schema);

      // find schema node
      for (piter= top_nodes.begin(); piter != top_nodes.end(); ++piter)
      {
        prow= *piter;
        
        if (prow[_sel_columns._object] == schema)
        {
          Gtk::TreeModel::Children sub_nodes= prow.children();
          Glib::ustring tname(table->table);
          // find table node
          for (citer= sub_nodes.begin(); citer != sub_nodes.end(); ++citer)
          {
            crow= *citer;
            
            if (crow[_sel_columns._object] == tname)
            {
              crow[_sel_columns._state]= MGCellRendererTristate::ON;
              break;
            }
          }
          // if all nodes are checked, check parent too
          if (seen[key]==sub_nodes.size())
          {
            prow[_sel_columns._state]= MGCellRendererTristate::ON;
          }
          else if (seen[key] > 0) // at least one checked
          {
            prow[_sel_columns._state]= MGCellRendererTristate::MAYBE;
          }
          else
          {
            prow[_sel_columns._state]= MGCellRendererTristate::OFF;
          }
          break;
        }
      }
    }
  }

  update_sensitivity(true);
}



MYX_BACKUP_CONTENT *MABackupPanel::gather_selected_tables()
{
  Gtk::TreeIter siter, titer;
  int alloced= 8, count= 0;
  MYX_BACKUP_CONTENT *content;
  MYX_BACKUP_TABLE *tables= (MYX_BACKUP_TABLE*)g_malloc0(alloced*sizeof(MYX_BACKUP_TABLE));

  for (siter= _selection_store->children().begin();
       siter != _selection_store->children().end();
       ++siter)
  {
    Gtk::TreeModel::Row row= *siter;

    for (titer= row.children().begin(); titer != row.children().end(); ++titer)
    {
      Gtk::TreeModel::Row rrow= *titer;
      Glib::ustring str;

      if (rrow[_sel_columns._state] == MGCellRendererTristate::ON)
      {
        if (count == alloced)
        {
          alloced += 10;
          tables= (MYX_BACKUP_TABLE*)g_realloc(tables, alloced*sizeof(MYX_BACKUP_TABLE));
        }

        tables[count].catalog= g_strdup("def");//XXX
        str= row[_sel_columns._object];
        tables[count].schema= g_strdup(str.c_str());
        str= rrow[_sel_columns._object];
        tables[count].table= g_strdup(str.c_str());
        count++;
      }
    }
  }

  content= (MYX_BACKUP_CONTENT*)g_malloc0(sizeof(MYX_BACKUP_CONTENT));
  content->tables_num= count;
  content->tables= tables;

  return content;
}


MYX_BACKUP_PROFILE *MABackupPanel::gather_profile()
{
  MYX_BACKUP_PROFILE *profile;
  Glib::ustring str;
  
  profile= (MYX_BACKUP_PROFILE*)g_malloc0(sizeof(MYX_BACKUP_PROFILE));

  profile->profile_name= g_strdup(get_entry("name_entry")->get_text().c_str());

  str= get_entry("target_entry")->get_text();
  if (str.size()>1 && str[str.size()-1]!='/')
      str += "/";
  profile->target_directory= g_strdup(str.c_str());

  profile->filename_prefix= g_strdup(get_entry("prefix_entry")->get_text().c_str());

  int flags= 0;
  for (int i= 1; i <= 8; i++)
  {
    Gtk::CheckButton *check;
    char buf[32];
    sprintf(buf, "option%i_check", i);
    check= (Gtk::CheckButton*)get_widget(buf);
    if (check->get_active())
        flags |= backup_options[i-1];
  }
  if (get_toggle("lockall_radio")->get_active())
    flags |= MYX_B_LOCK_ALL_TABLES; 
  else if (get_toggle("singletrans_radio")->get_active())
    flags |= MYX_B_SINGLE_TRANSACTION;
  
  if (get_toggle("backup_all_check"))
    flags |= MYX_B_COMPLETE_SCHEMATAS;


  profile->options= flags;

  profile->backup_type= MYX_BT_SQL_SCRIPT;
  
  if (!(flags & MYX_B_COMPLETE_SCHEMATAS))
      profile->backup_content= gather_selected_tables();

  
  return profile;
}


void MABackupPanel::save_profile()
{
  MYX_BACKUP_PROFILE *profile;

  // gather data
  profile= gather_profile();
  
  // check data
  if (!*profile->profile_name)
  {
    Gtk::MessageDialog dlg(*_app->window(),
                           _("Please supply a name for this profile"),
                           Gtk::MESSAGE_ERROR,
                           Gtk::BUTTONS_OK,
                           true);
    dlg.run();
  }
  else if (strchr(profile->profile_name, '/'))
  {
    Gtk::MessageDialog dlg(*_app->window(),
                           _("Profile name cannot contain character '/'"),
                           Gtk::MESSAGE_ERROR,
                           Gtk::BUTTONS_OK,
                           true);
    dlg.run();
  }
  else
  {
    bool renamed= false;
    std::string profile_dir= _file_browser->get_directory()+"/";
    Gtk::TreeIter iter= _file_browser->get_selected();

    // delete old profile
    if (iter)
    {
      std::string name;
      _file_browser->get_row_object(iter, name);
      _file_browser->remove(name);
      renamed= true;
    }

    // save
    _profile_name= profile->profile_name;
    
    if (myx_save_profile(std::string(_profile_name+PROFILE_SUFFIX).c_str(),
                         profile_dir.c_str(),
                         profile) < 0)
    {
      g_warning("could not save profile '%s'", profile->profile_name);
    }

    _file_browser->refresh();

    _dirty= false;
    
    if (renamed)
    {
      Gtk::TreeIter node;
      if (_file_browser->find_node(profile->profile_name, _file_browser->_columns._name, node))
        _file_browser->set_selection(node);
    }
  }
  
  myx_free_profile(profile);
}


void MABackupPanel::row_toggled(const Glib::ustring &path)
{
  Gtk::TreeIter iter= _selection_store->get_iter(path);
  Gtk::TreeModel::Row row= *iter;
  
  // toggled the schema
  if (row[_sel_columns._what] == SelectionColumns::CSchema)
  {
    Gtk::TreeModel::Children ch= row->children();

    if (row[_sel_columns._state] != MGCellRendererTristate::ON)
    {
      unsigned long total_rows= 0, total_length= 0;
      // schema was checked: check all tables
      for (Gtk::TreeIter it= ch.begin(); it != ch.end(); ++it)
      {
        Gtk::TreeModel::Row r= *it;
        
        r[_sel_columns._state]= MGCellRendererTristate::ON;

        total_rows+= r[_sel_columns._rows];
        total_length+= r[_sel_columns._lengthn];
      }
      row[_sel_columns._state]= MGCellRendererTristate::ON;
      row[_sel_columns._rows]= total_rows;
      row[_sel_columns._lengthn]= total_length;
      row[_sel_columns._length]= format_value(total_length);
    }
    else
    {
      // schema was unchecked: unselect all tables
      for (Gtk::TreeIter it= ch.begin(); it != ch.end(); ++it)
      {
        Gtk::TreeModel::Row r= *it;
        
        r[_sel_columns._state]= MGCellRendererTristate::OFF;
      }
      row[_sel_columns._state]= MGCellRendererTristate::OFF;
      row[_sel_columns._rows]= 0;
      row[_sel_columns._lengthn]= 0;
      row[_sel_columns._length]= "0";
    }
  }
  else
  {
    Gtk::TreeIter par= row.parent();
    Gtk::TreeModel::Row rpar= *par;
    unsigned long tmp;

    if (row[_sel_columns._state] != MGCellRendererTristate::ON)
    {
      // table was checked
      row[_sel_columns._state]= MGCellRendererTristate::ON;

      // update totals
      tmp= rpar[_sel_columns._rows];
      tmp+= row[_sel_columns._rows];
      rpar[_sel_columns._rows]= tmp;

      tmp= rpar[_sel_columns._lengthn];
      tmp+= row[_sel_columns._lengthn];
      rpar[_sel_columns._lengthn]= tmp;
      rpar[_sel_columns._length]= format_value(tmp);
    }
    else
    {
      // table was unchecked
      row[_sel_columns._state]= MGCellRendererTristate::OFF;

      // update totals
      tmp= rpar[_sel_columns._rows];
      tmp-= row[_sel_columns._rows];
      rpar[_sel_columns._rows]= tmp;

      tmp= rpar[_sel_columns._lengthn];
      tmp-= row[_sel_columns._lengthn];
      rpar[_sel_columns._lengthn]= tmp;
      rpar[_sel_columns._length]= format_value(tmp);
    }
    
    int total= 0, total_checked= 0;
    Gtk::TreeIter it;
    // update parent row according to the state of the children
    for (it= rpar.children().begin(); 
         it != rpar.children().end() && (total == total_checked || total_checked == 0);
         ++it)
    {
      Gtk::TreeModel::Row r= *it;
      total++;
      if (r[_sel_columns._state] == MGCellRendererTristate::ON)
        total_checked++;
    }
    
    if (total == total_checked)
      rpar[_sel_columns._state]= MGCellRendererTristate::ON;
    else if (total_checked == 0)
      rpar[_sel_columns._state]= MGCellRendererTristate::OFF;
    else
      rpar[_sel_columns._state]= MGCellRendererTristate::MAYBE;
  }
  
  update_button_bar_sensitivity();
  set_dirty();
}


void MABackupPanel::set_dirty()
{
  _dirty= true;
  
  update_button_bar_sensitivity();
}


int MABackupPanel::backup_progress_cb(const char *table,
                                      int num_tables_total,
                                      int num_tables,
                                      int num_rows_total,
                                      int num_rows,
                                      void *user_data)
{
  ProgressData *data= (ProgressData*)user_data;

  if (!data->ready)
  {
    data->table= table;
    data->num_tables_total= num_tables_total;
    data->num_tables= num_tables;
    data->num_rows_total= num_rows_total;
    data->num_rows= num_rows;
    data->ready= true;
  }
  return backup_aborted ? -1 : 0;
}


bool MABackupPanel::update_backup_progress(void *data)
{
  ProgressData *p= (ProgressData*)data;

  if (p->ready)
  {
    _progress_label->set_text(p->table);

    _row_progress->set_fraction((double)p->num_rows/p->num_rows_total);
    _row_progress->set_text(ufmt(_("row %i of %i"), p->num_rows, p->num_rows_total));
  
    _table_progress->set_fraction((double)p->num_tables/p->num_tables_total);
    _table_progress->set_text(ufmt(_("table %i of %i"), p->num_tables, p->num_tables_total));
    
    p->ready= false;
  }

  return true;
}


void *MABackupPanel::backup_thread(void *data)
{
  BackupThreadArgument *arg= (BackupThreadArgument*)data;
  MYX_BACKUP_ERROR err;
  MInstanceInfo *inst= arg->me->_data->get_instance();

  inst->lock_mysql(true);

  err= myx_make_backup_with_profile(inst->get_mysql(), arg->profile, 
                                    arg->path.c_str(),
                                    1, backup_progress_cb, &arg->pdata);

  inst->lock_mysql(false);

  return (void*)err;
}


void MABackupPanel::stop_backup()
{
  backup_aborted= true;
}


void MABackupPanel::perform_backup()
{
  BackupThreadArgument arg;
  MDispatcher disp;
  char suffix[128];
  time_t now= time(NULL);
  Gtk::Dialog *dlg= (Gtk::Dialog*)_progress_xml->get_widget("progress_dialog");
  MYX_BACKUP_ERROR err;

  _progress_label->set_text(_("Preparing..."));
  _table_progress->set_fraction(0);
  _row_progress->set_fraction(0);

  strftime(suffix, sizeof(suffix), "_%F_%T.sql", localtime(&now));

  arg.me= this;
  {
    Gtk::FileSelection fsel(_("Select Backup File Name"));
    fsel.set_filename(_profile_name+std::string(suffix));

    if (fsel.run() != Gtk::RESPONSE_OK)
      return;

    arg.path= fsel.get_filename();
  }
  _app->push_status(_("Performing backup..."));

  arg.profile= gather_profile();
  arg.pdata.ready= false;

  backup_aborted= false;

  {    
    SigC::Connection con= Glib::signal_timeout().connect(SigC::bind<void*>
                                                         (SigC::slot(*this, &MABackupPanel::update_backup_progress),&arg.pdata),
                                                         50);
    dlg->show();

    err= (MYX_BACKUP_ERROR)(int)disp.dispatch(backup_thread, &arg);

    dlg->hide();

    con.disconnect();
  }

  myx_free_profile(arg.profile);
  
  _app->pop_status();

  Glib::ustring msg;

  switch (err)
  {
  case MYX_BACKUP_MYSQL_ERROR:
    msg= _("MySQL error during backup.");
    break;
  case MYX_BACKUP_CANT_OPEN_FILE:
    msg= _("Could not open file for writing backup.");
    break;
  case MYX_BACKUP_ILLEGAL_OPTION:
    msg= _("Illegal option.");
    break;
  case MYX_BACKUP_PCRE_ERROR:
    msg= _("Internal error in PCRE call.");
    break;
  case MYX_BACKUP_MALLOC_FAILED:
    msg= _("Memory allocation failed during backup.");
    break;
  case MYX_BACKUP_OUTPUTDEVICE_FULL:
    msg= _("Out of disk space during backup write.");
    break;
  case MYX_BACKUP_NO_ERROR:
  default:
    {
      if (backup_aborted)
        _app->set_status(_("Backup cancelled."));
      else
        _app->set_status(_("Backup finished."));

      Gtk::MessageDialog dlg(*_app->window(),
                             ufmt(_("Backup written to file '%s'."),
                                  arg.path.c_str()),
                             Gtk::MESSAGE_INFO,
                             Gtk::BUTTONS_OK,
                             true);
      dlg.run();
    }
  }
  
  if (!msg.empty())
  {
    Gtk::MessageDialog dlg(*_app->window(), _("Error during backup.\n")+msg,
                           Gtk::MESSAGE_ERROR,
                           Gtk::BUTTONS_OK,
                           true);
    dlg.run();
  }
}


void MABackupPanel::schedule_type_changed()
{
  int i= ((Gtk::OptionMenu*)get_widget("schedule_menu"))->get_history();

  ((Gtk::Notebook*)get_widget("schedule_note"))->set_current_page(i);
  get_widget("schedule_frame")->set_sensitive(i > 0);
}


void MABackupPanel::gather_schedule(MABackupPanel::Schedule &schedule)
{
  switch (((Gtk::OptionMenu*)get_widget("schedule_menu"))->get_history())
  {
  case 0:
    schedule.type= Schedule::SNone;
    break;
  case 1:
    schedule.type= Schedule::SPeriodically;
    schedule.ptype= (Schedule::PType)((Gtk::OptionMenu*)get_widget("period_menu"))->get_history();
    schedule.period_length= get_spin("period_spin")->get_value_as_int();
    schedule.time= get_spin("periodh_spin")->get_text()+":"+get_spin("periodm_spin")->get_text();
    break;
  case 2:
    schedule.type= Schedule::SWeekly;
    schedule.wday= 0;
    for (int i= 1; i <= 7; i++)
    {
      if (get_toggle("week"+tostr(i)+"_check")->get_active())
      {
        schedule.wday= i;
        break;
      }
    }
    schedule.time= get_spin("weekh_spin")->get_text()+":"+get_spin("weekm_spin")->get_text();
    break;
  case 3:
    schedule.type= Schedule::SMonthly;
    schedule.mday= ((Gtk::OptionMenu*)get_widget("month_menu"))->get_history()+1;
    schedule.time= get_spin("monthh_spin")->get_text()+":"+get_spin("monthm_spin")->get_text();
    break;
  }
}


std::string MABackupPanel::make_cron_line(Schedule &schedule)
{
  std::string str;
  
  switch (schedule.type)
  {
  default:
    break;
  }
  
  return str;
}


bool MABackupPanel::setup_schedule(MABackupPanel::Schedule &schedule)
{
  return false;
}


/**********************************************************************/

void MABackupPanel::schema_selected(MGBrowserList *sender,
                                    const Gtk::TreeIter &node)
{
  _selection_tree->get_selection()->unselect_all();

  update_button_sensitivity();
}


void MABackupPanel::profile_selected(MGBrowserList *sender,
                                     const Gtk::TreeIter &node)
{
  MYX_BACKUP_PROFILE *profile;
  MYX_ADMIN_LIB_ERROR err;
  std::string name;

  if (node)
  {
    std::string profile_dir= _file_browser->get_directory()+"/";

    _file_browser->get_row_object(node, name);
    name+= PROFILE_SUFFIX;
    
    profile= myx_load_profile(name.c_str(), profile_dir.c_str(), &err);
    if (profile)
    {
      show_profile(profile);

      myx_free_profile(profile);
    }
    else
    {
      show_adminlib_error(*_app->window(),
                          ufmt(_("Could not load backup profile '%s/%s'"),
                               profile_dir.c_str(), name.c_str()),
                          err);
    }
  }
  else
  {
    show_profile(0);
  }
}


/**********************************************************************/

MAPanel *create_backup_panel(MAdministrator *app, MDataInterface *data)
{
  return new MABackupPanel(app, data);
}
