/* Copyright (C) 2003, 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include "myadmin.h"
#include "MAdministrator.h"
#include "MAServerLogsPanel.h"
#include "MDataInterface.h"

#include <errno.h>
#include <math.h>

#include "myg_utils.h"
#include "myx_admin_public_interface.h"


#define PAGE_BLOCK_SIZE (1024*5)


MAServerLogsPanel::MAServerLogsPanel(MAdministrator *app, MDataInterface *data)
  : MAPanel(app, data)
{
}


void MAServerLogsPanel::setup_panel(LogType type)
{
  Gtk::TreeView *tree= _tree[type];
//  _menu[type]->set_sensitive(false);

  tree->append_column("", _columns._icon);
  tree->append_column(_("Time"), _columns._time);
  tree->append_column(_("Event"), _columns._text);

  tree->get_column(0)->set_fixed_width(30);
  tree->get_column(1)->set_resizable(true);
  tree->get_column(2)->set_resizable(true);

  _store[type]= Gtk::ListStore::create(_columns);
  tree->set_model(_store[type]);
}


bool MAServerLogsPanel::init()
{
  int i;
  static char *types[]={
    "error","binary","slow","general","innodb"
  };

  if (_xml)
    return true;

  if (!MAPanel::init_from_glade(GLADE_DATADIR"/"GLADE_SERVERLOGS_FILE, "panel_frame"))
    return false;

  for (i= 0; i < 5; i++)
  {
    char buffer[32];
    
    sprintf(buffer, "%s_label", types[i]);
    _label[i]= get_label(buffer);

    sprintf(buffer, "%s_frame", types[i]);
    _frame[i]= get_label(buffer);

    sprintf(buffer, "%s_tree", types[i]);
    _tree[i]= get_tree(buffer);

    _tree[i]->get_selection()->signal_changed().connect(SigC::bind<LogType>(SigC::slot(*this,&MAServerLogsPanel::selected_event),(LogType)i));

//    sprintf(buffer, "%s_option", types[i]);
//    _menu[i]= (Gtk::OptionMenu*)get_widget(buffer);
    sprintf(buffer, "%s_text", types[i]);
    _textv[i]= (Gtk::TextView*)get_widget(buffer);
    _text[i]= _textv[i]->get_buffer();

    sprintf(buffer, "%s_scrollbar", types[i]);
    _sbar[i]= (Gtk::Scrollbar*)get_widget(buffer);

    _sbar[i]->signal_value_changed().connect(SigC::bind<LogType>(SigC::slot(*this,&MAServerLogsPanel::scroll_log),(LogType)i));

    sprintf(buffer, "%s_save", types[i]);
    get_button(buffer)->signal_clicked().connect(SigC::bind<LogType>(SigC::slot(*this,&MAServerLogsPanel::save_log),(LogType)i));
  }

  _start_icon= _app->load_pixbuf("service_start.png");
  _stop_icon= _app->load_pixbuf("service_stop.png");

  _warning_icon= _app->load_pixbuf("16x16_warning.png");
  _error_icon= _app->load_pixbuf("16x16_error.png");
  
  setup_panel(LError);
  setup_panel(LBinary);
  setup_panel(LSlow);
  setup_panel(LGeneral);
  setup_panel(LInnoDB);

  return true;
}


bool MAServerLogsPanel::before_show()
{
  process_logs(LError);
  process_logs(LInnoDB);
  process_logs(LGeneral);

  return true;
}


bool MAServerLogsPanel::before_hide()
{

  return true;
}


void MAServerLogsPanel::selected_event(LogType type)
{
  Gtk::TreeIter iter= _tree[type]->get_selection()->get_selected();
  
  if (iter)
  {
    Gtk::TreeModel::Row row= *iter;
    int line= row[_columns._line];
    Gtk::TextIter lit= _text[type]->get_iter_at_line(line);
    
    _textv[type]->scroll_to_iter(lit, 0.0);
    //XXX select line
  }
}


void MAServerLogsPanel::scroll_log(LogType type)
{
  double val= floor(_sbar[type]->get_value());
  _sbar[type]->set_value(val);

  process_logs(type);
}


void MAServerLogsPanel::save_log(LogType type)
{
  Gtk::FileSelection fsel(_("Save Log"));

  if (fsel.run() != Gtk::RESPONSE_OK)
    return;
  
  std::string target= fsel.get_filename();
  std::string source;
  MInstanceInfo *inst= _data->get_instance();  

  switch (type)
  {
  case LError: source= inst->get_server_path(MInstanceInfo::PErrorLog); break;
  case LGeneral: source= inst->get_server_path(MInstanceInfo::PLogs) + "/mysql.log"; break;
  case LInnoDB: source= inst->get_server_path(MInstanceInfo::PInnoDBLog); break;
  default: break;
  }

  if (copy_file(source.c_str(), target.c_str()) < 0)
  {
    show_sys_error(*_app->window(),
                   ufmt(_("Could not save log file to '%s'"), target.c_str()),
                   errno);
  }
  else
  {
    _app->set_status(ufmt(_("Log file saved to '%s'"), target.c_str()));
  }
}


void MAServerLogsPanel::process_logs(LogType type)
{
  std::string file;
  int block;
  MYX_LOGFILE *logf;
  MInstanceInfo *inst= _data->get_instance();
  long total_size;

  switch (type)
  {
  case LError: file= inst->get_server_path(MInstanceInfo::PErrorLog); break;
  case LGeneral: file= inst->get_server_path(MInstanceInfo::PLogs); break;
  case LInnoDB: file= inst->get_server_path(MInstanceInfo::PInnoDBLog); break;
  default: break;
  }
  
  logf= myx_parse_error_log(file.c_str(), PAGE_BLOCK_SIZE, 
                            (int)_sbar[type]->get_value()+1, &block);
  if (logf)
  {    
    _frame[type]->set_sensitive(true);

    total_size= get_file_size(file.c_str());

    _sbar[type]->set_range(0, total_size/PAGE_BLOCK_SIZE);
    _sbar[type]->get_adjustment()->set_page_size(1);
    _sbar[type]->set_increments(1, 1);

    _store[type]->clear();
    _text[type]->assign("");

    for (unsigned int i= 0; i < logf->events_num; i++)
    {
      MYX_LOGFILE_EVENT *event= logf->events + i;
      Gtk::TreeIter iter = _store[type]->append();
      Gtk::TreeModel::Row row= *iter;
      
      switch (event->event_type)
      {
      case MYX_EVENT_START:
        row[_columns._icon]= _start_icon;
        row[_columns._text]= _("Server startup");
        break;
      case MYX_EVENT_END:
        row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Server shutdown");
        break;
      case MYX_EVENT_ERROR:
        row[_columns._icon]= _error_icon;
        row[_columns._text]= _("Error");
        break;
      case MYX_EVENT_INNODB_START:
        row[_columns._icon]= _start_icon;
        row[_columns._text]= _("InnoDB startup");
        break;
      case MYX_EVENT_INNODB_SHUTDOWN:
        row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("InnoDB shutdown");
        break;
      case MYX_EVENT_FORCED_CLOSE_THREAD:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Thread closed");
        break;
      case MYX_EVENT_ABORT:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Abort");
        break;
      case MYX_EVENT_SELECT:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Select");
        break;
      case MYX_EVENT_INIT:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Init DB");
        break;
      case MYX_EVENT_CONNECT:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Connect");
        break;
      case MYX_EVENT_QUIT:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Quit");
        break;
      case MYX_EVENT_QUERY:
        //row[_columns._icon]= _stop_icon;
        row[_columns._text]= _("Query");
        break;
      }
      
      row[_columns._line]= event->line_no;

      char buffer[64];
      struct tm t;
      t.tm_year= event->date->tm_year;
      t.tm_mon= event->date->tm_mon;
      t.tm_mday= event->date->tm_mday;
      t.tm_hour= event->date->tm_hour;
      t.tm_min= event->date->tm_min;
      t.tm_sec= event->date->tm_sec;
      sprintf(buffer, "%i/%i/%i %i:%i:%i", 
              t.tm_mon, t.tm_mday, t.tm_year,
              t.tm_hour, t.tm_min, t.tm_sec);
      strptime(buffer,"%D %T",&t);
      strftime(buffer, sizeof(buffer), "%a %X %x", &t);
      row[_columns._time]= buffer;
    }
    
    for (unsigned int i= 0; i < logf->lines_num; i++)
    {
      Gtk::TextIter iter= _text[type]->end();

      _text[type]->insert(iter, Glib::ustring((char*)logf->lines[i])+"\n");
    }

    myx_free_logfile(logf);
  }
  else
  {
    _frame[type]->set_sensitive(false);
//    g_warning("could not open log file '%s'", file.c_str());
  }
}




MAPanel *create_server_logs_panel(MAdministrator *app, MDataInterface *data)
{
  return new MAServerLogsPanel(app, data);
}
