/* Copyright (C) 2003, 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include "myadmin.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "MGGladeXML.h"

#include "MAdministrator.h"
#include "MAPanel.h"
#include "MDataInterface.h"

#include "myg_gtk_utils.h"

#include "MGConnectDialog.h"
#include "MGPreferencesEditor.h"

#include "MAServerInformationPanel.h"
#include "MAServiceControlPanel.h"
#include "MAStartupParametersPanel.h"
#include "MAServerConnectionsPanel.h"
#include "MAUserAdministrationPanel.h"
#include "MAHealthPanel.h"
#include "MAServerLogsPanel.h"
#include "MABackupPanel.h"
#include "MARestorePanel.h"
#include "MAReplicationStatusPanel.h"
#include "MACatalogsPanel.h"

#include "MADummyPanel.h"


static struct sidebar_item {
  char *icon;
  char *icon_gray;
  Glib::ustring label;
  char *menu_item;
  MAPanel*(*instantiate)(MAdministrator*,MDataInterface*);
} sidebar_item_list[]=
{
  { "server_info.png", "server_info_gray.png", 
      N_("Server Information"), "server_information1",
      create_server_information_panel },
  { "service_config.png", "service_config_gray.png",
      N_("Service Control"), "service_control1",
      create_service_control_panel },
  { "startup_variables.png", "startup_variables_gray.png",
      N_("Startup Variables"), "startup_parameters1",
      create_startup_parameters_panel },
  { "user_admin.png", "user_admin_gray.png",
      N_("User Administration"), "user_administration1",
      create_user_administration_panel },
  { "server_connections.png", "server_connections_gray.png",
      N_("Server Connections"), "server_connections1",
      create_server_connections_panel },
  { "health.png", "health_gray.png",
      N_("Health"), "health1",
      create_health_panel },
  { "server_logs.png", "server_logs_gray.png",
      N_("Server Logs"), "server_logs1",
      create_server_logs_panel },
  { "backup.png", "backup_gray.png",
      N_("Backup"), "backup1",
      create_backup_panel },
  { "restore.png", "restore_gray.png",
      N_("Restore Backup"), "restore1",
      create_restore_panel },
  { "replication.png", "replication_gray.png",
      N_("Replication Status"), "replication_status1",
      create_replication_status_panel },
  { "catalogs.png", "catalogs_gray.png",
      N_("Catalogs"), "catalogs1",
      create_catalogs_panel },
};


void MAdministrator::populate_sidebar()
{  
  unsigned int i;
  _side_store= Gtk::TreeStore::create(_sidebar_columns);
  bool is_local, is_connected;
  Gdk::Color normal= _tree->get_style()->get_fg(Gtk::STATE_NORMAL);
  Gdk::Color gray= _tree->get_style()->get_fg(Gtk::STATE_INSENSITIVE);

  is_connected= _instance->is_connected();
  is_local= _instance->is_local();

  for (i= 0; i < sizeof(sidebar_item_list)/sizeof(struct sidebar_item); i++)
  {
    Gtk::TreeModel::iterator iter= _side_store->append();
    Gtk::TreeModel::Row row= *iter;
    MAPanel *panel= 0;
    bool is_active= true;

    panel= (*sidebar_item_list[i].instantiate)(this, _dataif);

    if (panel->is_local_only() && !is_local)
      is_active= false;
    if (panel->needs_connection() && !is_connected)
      is_active= false;

    row[_sidebar_columns._panel]= panel;

    // set menu callback
    Gtk::MenuItem *item= (Gtk::MenuItem*)get_widget(sidebar_item_list[i].menu_item);
    item->signal_activate().connect(SigC::bind<int>(SigC::slot(*this,&MAdministrator::switch_panel_num),i));

    // set icon
    try
    {
      row[_sidebar_columns._icon]= load_pixbuf(is_active
                                               ?sidebar_item_list[i].icon
                                               :sidebar_item_list[i].icon_gray);
    } 
    catch (...) 
    {
      g_message("cant load icon %s\n", sidebar_item_list[i].icon);
    }

    row[_sidebar_columns._text]= sidebar_item_list[i].label;

    row[_sidebar_columns._color]= is_active ? normal : gray;
  }

  _tree->set_model(_side_store);
}



void MAdministrator::setup_sidebar()
{
  Glib::RefPtr<Gtk::TreeSelection> select;

  _tree->append_column("", _sidebar_columns._icon);

  _tree->append_column("", _sidebar_columns._text);
  Gtk::TreeViewColumn *column= _tree->get_column(1);
  column->add_attribute(((Gtk::CellRendererText*)column->get_first_cell_renderer())->property_foreground_gdk(),
                       _sidebar_columns._color);

  populate_sidebar();

  select= _tree->get_selection();
  select->signal_changed().connect(SigC::slot(*this, &MAdministrator::tree_selection_changed_cb));
  select->set_mode(Gtk::SELECTION_BROWSE);
}


Gtk::Widget *MAdministrator::get_widget(const Glib::ustring &name)
{
  Gtk::Widget *w= _xml->get_widget(name);

  if (!w)
  {
    g_critical("request for non-existent widget %s in glade XML", name.c_str());
    g_assert_not_reached();
  }

  return w;
}


MAdministrator::MAdministrator()
  : _xml(0), _about_xml(0), _prefs_editor(0), _tmp_status(0), _instance(0), 
    _dataif(0), _changing_panel(false)
{
  _current_panel= NULL;
}


MAdministrator::~MAdministrator()
{
  delete _dataif;
  delete _xml;
}



bool MAdministrator::init(MInstanceInfo *inst)
{
  _instance= inst;
  _dataif= new MDataInterface(this, inst);
  
  _xml= new MGGladeXML(GLADE_DATADIR"/"GLADE_MAINWINDOW_FILE);
  if (!_xml)
  {
    Gtk::MessageDialog dlg(ufmt(_("The user interface file (%s) could not be loaded.\n"
                                  "%s was probably not installed correctly."),
                                APP_DATADIR"/"GLADE_MAINWINDOW_FILE,
                                PACKAGE),
                           Gtk::MESSAGE_ERROR,
                           Gtk::BUTTONS_OK,
                           true);
    dlg.run();
    return false;
  }

  _win= (Gtk::Window*)get_widget("main_window");
  _tree= (Gtk::TreeView*)get_widget("sidebar_tree");

  _win->signal_delete_event().connect(SigC::slot(*this,&MAdministrator::close_cb));
  ((Gtk::MenuItem*)get_widget("new_connection1"))->signal_activate().connect(SigC::slot(*this,&MAdministrator::new_connection_cb));
  ((Gtk::MenuItem*)get_widget("close_connection1"))->signal_activate().connect(SigC::slot(*this,&MAdministrator::close_connection_cb));
  ((Gtk::MenuItem*)get_widget("quit1"))->signal_activate().connect(SigC::slot(*this,&MAdministrator::menu_quit_cb));
  ((Gtk::MenuItem*)get_widget("about1"))->signal_activate().connect(SigC::slot(*this,&MAdministrator::menu_about_cb));

  ((Gtk::MenuItem*)get_widget("preferences1"))->signal_activate().connect(SigC::slot(*this,&MAdministrator::menu_preferences_cb));
  
  _progressbar= (Gtk::ProgressBar*)get_widget("status_progress");

  _dummy_panel= new MADummyPanel(this, _dataif);
  
  setup_sidebar();

  ((Gtk::Button*)_xml->get_widget("status_button"))->signal_clicked().connect(SigC::slot(*this, &MAdministrator::cancel_button_clicked_cb));

  if (!_transl.open(APP_DATADIR"/mysqlx_translations_general.xml",
                    APP_DATADIR"/mysqlx_translations_administrator.xml",
                    "en"))
  {
    g_warning(_("Could not open translation files"));
  }

  _instance->signal_disconnect().connect(SigC::slot(*this,&MAdministrator::disconnect_db_handler));

  // set window title
  if (_instance->is_connected())
  {
    MYX::UserConnection conn= _instance->get_connection_data();
    if (_instance->is_socket_connection())
      _win->set_title(ufmt(MY_NAME" %s@%s via socket",
                           conn.username.c_str(), conn.hostname.c_str()));
    else
      _win->set_title(ufmt(MY_NAME" %s@%s:%i",
                           conn.username.c_str(), conn.hostname.c_str(), conn.port));
  }
  else
  {
    _win->set_title(ufmt(MY_NAME" %s", _("(not connected)")));
  }

  // setup icons
  std::list<Glib::RefPtr<Gdk::Pixbuf> > icons;
  icons.push_back(Gdk::Pixbuf::create_from_file(DATADIR"/MySQLIcon_Admin_16x16.png"));
  icons.push_back(Gdk::Pixbuf::create_from_file(DATADIR"/MySQLIcon_Admin_32x32.png")),
  icons.push_back(Gdk::Pixbuf::create_from_file(DATADIR"/MySQLIcon_Admin_48x48.png"));
  _win->set_icon_list(icons);

  return true;
}


Glib::RefPtr<Gdk::Pixbuf> MAdministrator::load_pixbuf(const std::string &fname)
{
  Glib::RefPtr<Gdk::Pixbuf> pix;

  if (_pixbuf_cache.find(fname) == _pixbuf_cache.end())
  {
    try {
      pix= Gdk::Pixbuf::create_from_file(APP_DATADIR"/"+fname);
    } 
    catch (...)
    {
      g_error("Error loading image '%s'", std::string(APP_DATADIR"/"+fname).c_str());
    }
    _pixbuf_cache[fname]= pix;
  }

  pix= _pixbuf_cache[fname];
  
  if (pix)
    pix->reference();
  return pix;
}


MInstanceInfo *MAdministrator::get_instance() const
{
  g_assert(_instance != 0);

  return _instance; 
}


void MAdministrator::show()
{
  _win->show();

  // calculate the min size of the sidebar to fit all rows
  Gdk::Rectangle rect;
  _tree->get_background_area(Gtk::TreePath(_side_store->children().begin()),
                             *_tree->get_column(0), rect);

  _tree->set_size_request(-1, rect.get_height()*_side_store->children().size());

  _win->maximize();
}


void MAdministrator::disconnect_db_handler()
{
  update_sidebar(true);
  
  if (_current_panel && _current_panel->needs_connection())
  {
    ((MADummyPanel*)_dummy_panel)->set_connection_text();
    switch_panel(_dummy_panel);
  }
}


void MAdministrator::update_sidebar(bool disconnecting)
{
  Gtk::TreeIter iter;
  Gtk::TreeRow row;
  Gdk::Color normal= _tree->get_style()->get_fg(Gtk::STATE_NORMAL);
  Gdk::Color gray= _tree->get_style()->get_fg(Gtk::STATE_INSENSITIVE);

  int i= 0;
  for (iter= _side_store->children().begin();
       iter != _side_store->children().end();
       ++iter, i++)
  {
    row= *iter;
    MAPanel *panel= row[_sidebar_columns._panel];
    
    if ((panel->needs_connection() && (!_instance->is_connected() || disconnecting))
        || (panel->is_local_only() && !_instance->is_local()))
    {
      row[_sidebar_columns._icon]= load_pixbuf(sidebar_item_list[i].icon_gray);
      row[_sidebar_columns._color]= gray;
    }
    else
    {
      row[_sidebar_columns._icon]= load_pixbuf(sidebar_item_list[i].icon);
      row[_sidebar_columns._color]= normal;
    }
  }
}


void MAdministrator::quit()
{
  Gtk::TreeIter iter;
  bool ok= true;

  for (iter= _side_store->children().begin();
       iter != _side_store->children().end();
       ++iter)
  {
    Gtk::TreeModel::Row row= *iter;
    MAPanel *panel= row[_sidebar_columns._panel];
    
    if (panel)
    {
      if (!panel->before_quit())
      {
        ok= false;
        break;
      }
    }
  }
  
  if (ok)
  {
    prefs.save();

    Gtk::Main::instance()->quit();
  }
}


Gtk::Container *MAdministrator::parent_frame()
{
  return (Gtk::Container*)get_widget("parent_frame");
}


Gtk::Window *MAdministrator::window()
{
  return _win;
}


void MAdministrator::switch_panel(const Glib::ustring &name)
{
  Gtk::TreeModel::Children children= _side_store->children();

  for (Gtk::TreeIter iter= children.begin(); iter != children.end(); ++iter)
  {
    Gtk::TreeModel::Row row= *iter;
    if (row[_sidebar_columns._text] == name)
    {
      _tree->get_selection()->select(iter);
      break;
    }
  }
}


void MAdministrator::switch_panel_num(int index)
{
  if (_changing_panel)
      return;

  Gtk::TreeModel::Children children= _side_store->children();
  
  for (Gtk::TreeIter iter= children.begin(); iter != children.end(); ++iter)
  {
    Gtk::TreeModel::Row row= *iter;
    if (index--==0)
    {
      if (_tree->get_selection()->get_selected()!=iter)
        _tree->get_selection()->select(iter);
      break;
    }
  }
}


void MAdministrator::switch_panel(MAPanel *panel)
{
  bool changed= false;

  if (panel->is_local_only() && !_instance->is_local())
  {
    panel= _dummy_panel;
    ((MADummyPanel*)_dummy_panel)->set_local_text();
  }
  if (panel->needs_connection() && !_instance->is_connected())
  {
    panel= _dummy_panel;
    ((MADummyPanel*)_dummy_panel)->set_connection_text();
  }
  
  if (panel == _current_panel)
    return;

  if (panel->init())
  {
    if (_current_panel)
    {
      if (_current_panel->before_hide() && panel->before_show())
      {
        _current_panel->hide();
        panel->show();

        _current_panel= panel;
        changed= true;
      }
    }
    else
    {
      if (panel->before_show())
      {
        panel->show();
        _current_panel= panel;
        
        changed= true;
      }
    }
  }
  
  // select the corresponding menu item
  Gtk::TreeModel::Children ch= _side_store->children();
  int i= 0;
  for (Gtk::TreeIter it= ch.begin(); it != ch.end(); ++it, ++i)
  {
    if (it == _tree->get_selection()->get_selected())
      break;
  }
  _changing_panel= true;
  if (!((Gtk::RadioMenuItem*)get_widget(sidebar_item_list[i].menu_item))->get_active())
      ((Gtk::RadioMenuItem*)get_widget(sidebar_item_list[i].menu_item))->activate();
  _changing_panel= false;

  if (!changed)
  {
    //XXX 
    // undo tree selection change
  }
}


void MAdministrator::add_side_panel(Gtk::Widget *panel)
{
  Gtk::Box *box= (Gtk::Box*)get_widget("sidebar_box");
  Gtk::Widget *cnt= _tree->get_parent();

  // readd the tree non-expanding
  cnt->reference();
  box->remove(*cnt);
  box->pack_start(*cnt, false, true);

  box->pack_start(*panel, true, true, 0);
  panel->show();
}


void MAdministrator::remove_side_panel(Gtk::Widget *panel)
{
  Gtk::Box *box= (Gtk::Box*)get_widget("sidebar_box");
  Gtk::Widget *cnt= _tree->get_parent();

  // readd the tree expanding
  cnt->reference();
  box->remove(*cnt);
  box->pack_start(*cnt, true, true);

  panel->reference();
  panel->hide();
  box->remove(*panel);
}


Gtk::MenuItem *MAdministrator::add_section_menu(Gtk::Menu *menu, const Glib::ustring &label)
{
  Gtk::MenuItem *item;

  item= new Gtk::MenuItem(label, true);
  item->set_submenu(*menu);

  ((Gtk::MenuBar*)get_widget("main_menu"))->insert(*item, 3);

  return item;
}



void MAdministrator::push_status(const Glib::ustring &text)
{
  Gtk::Statusbar *s= (Gtk::Statusbar*)get_widget("status_bar");

  if (_tmp_status)
  {
    s->remove_message(_tmp_status);
    _tmp_status= 0;
  }
  s->push(text);
}


void MAdministrator::pop_status()
{
  Gtk::Statusbar *s= (Gtk::Statusbar*)get_widget("status_bar");

  s->pop();
}


void MAdministrator::set_status(const Glib::ustring &text)
{
  Gtk::Statusbar *s= (Gtk::Statusbar*)get_widget("status_bar");

  _tmp_status= s->push(text);
}


void MAdministrator::set_busy(const Glib::ustring &message)
{
  Gtk::ProgressBar *pbar;
  Gtk::Button *btn;
  Gtk::Paned *paned;
  Gtk::MenuBar *menu;

  pbar= (Gtk::ProgressBar*)get_widget("status_progress");
  btn= (Gtk::Button*)get_widget("status_button");
  paned= (Gtk::Paned*)get_widget("main_paned");
  menu= (Gtk::MenuBar*)get_widget("main_menu");

  if (!message.empty())
  {
    pbar->show();
    btn->show();
    menu->set_sensitive(false);
    paned->set_sensitive(false);

    pbar->set_text(_("Please wait..."));
    
    push_status(message);
  }
  else
  {
    pbar->set_text("");
    pbar->set_fraction(0.0);
    pop_status();

    pbar->hide();
    btn->hide();
    menu->set_sensitive(true);
    paned->set_sensitive(true);
  }
}


void MAdministrator::set_busy_progress(bool start)
{
  if (start)
    _pulse_conn= Glib::signal_timeout().connect(SigC::slot(*this, &MAdministrator::pulse_progress),
                                                10000);
  else
    _pulse_conn.disconnect();
}


bool MAdministrator::pulse_progress()
{
  _progressbar->pulse();
  return true;
}


/**********************************************************************/


void MAdministrator::close_about()
{
  _about_xml->get_widget("about_window")->hide();
}


void MAdministrator::new_connection_cb()
{
  MGConnectDialog connect_dlg(&prefs);

  if (connect_dlg.run("MySQL Administrator", 0))
  {
    if (_instance->is_connected())
    {
      _instance->disconnect();
      _win->set_title(ufmt(MY_NAME" %s", _("(not connected)")));
    }

    MYX::UserConnection conn= connect_dlg.get_user_connection();

    _instance->set_connection(connect_dlg.get_connection(), conn);

    if (_instance->is_socket_connection())
      _win->set_title(ufmt(MY_NAME" %s@%s via socket",
                           conn.username.c_str(), conn.hostname.c_str()));
    else
      _win->set_title(ufmt(MY_NAME" %s@%s:%i",
                           conn.username.c_str(), conn.hostname.c_str(), conn.port));

    update_sidebar();

    Gtk::TreeIter iter= _tree->get_selection()->get_selected();
    if (iter)
    {
      Gtk::TreeRow row= *iter;

      switch_panel(row[_sidebar_columns._panel]);
    }
  }
}


void MAdministrator::close_connection_cb()
{
  if (_instance->is_connected())
  {
    _instance->disconnect();
    _win->set_title(ufmt(MY_NAME" %s", _("(not connected)")));
  }
}


bool MAdministrator::close_about_cb(GdkEventAny *ev)
{
  _about_xml->get_widget("about_window")->hide();
  return true;
}


void MAdministrator::menu_about_cb()
{
  if (!_about_xml)
  {
    _about_xml= new MGGladeXML(GLADE_DATADIR"/"GLADE_MAINWINDOW_FILE,
                               "about_window");
        
    ((Gtk::Button*)_about_xml->get_widget("close_button"))->signal_clicked().connect(SigC::slot(*this,&MAdministrator::close_about));
    ((Gtk::Window*)_about_xml->get_widget("about_window"))->signal_delete_event().connect(SigC::slot(*this,&MAdministrator::close_about_cb));
  }
  
  ((Gtk::Label*)_about_xml->get_widget("version_label"))->set_label(ufmt(_("Version %s")," "VERSION" ALPHA"));

  _about_xml->get_widget("about_window")->show();
}


void MAdministrator::menu_preferences_cb()
{
  if (!_prefs_editor)
    _prefs_editor= MGPreferencesEditor::create(&prefs);

  _prefs_editor->show();
}


void MAdministrator::menu_quit_cb()
{
  quit();
}


bool MAdministrator::close_cb(GdkEventAny *ev)
{
  quit();
  return true;
}


void MAdministrator::tree_selection_changed_cb()
{
  Gtk::TreeModel::iterator iter;

  iter= _tree->get_selection()->get_selected();
  if (iter)
  {
    Gtk::TreeModel::Row row= *iter;

    switch_panel(row[_sidebar_columns._panel]);
  }
}


void MAdministrator::cancel_button_clicked_cb()
{
  _instance->cancel_data_fetch();
}
