/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <myx_library.h>
#include <myx_xml_aux_functions.h>

typedef enum _kind { 
  APPLICATION_SPECIFIC_FILE = 1, GLOBAL_FILE=2
} FILE_KIND;

/* forward declarations */
static int parse_translation_file(MYX_TRANS *trans, const char *filename,const char *lang, MYX_LIB_ERROR *error_code,  FILE_KIND kind);
static void read_in_text(const xmlNodePtr text_node, MYX_TRANS_TEXT *text,const char *lang);
static void read_in_textgroup(const xmlNodePtr textgroup_node, MYX_TRANS_TEXTGROUP *group, const char *lang);
static void free_textgroup_content(MYX_TRANS_TEXTGROUP *group);
static void free_text_content(MYX_TRANS_TEXT *text);
static void sort_alphabetically(MYX_TRANS *trans);
static int compare_textgroups(const void *a, const void *b);
static int compare_transtexts(const void *a, const void *b);

/*
 *public functions definitions
 */

/* for testing...
int main()
{
  MYX_TRANS *t;
  MYX_LIB_ERROR error_code;
  const char *test;

  t = myx_init_trans("../xmlfiles/mysqlgui_translations_administrator.xml", "en", &error_code);

  if (t == NULL)
  {
    printf("Es trat folgender Fehler auf: %d\n", error_code);
    return 1;
  }
  test =  myx_t(t, "ResourceStrings", "FetchCatalogs", "son scheiss..");
  printf("%s\n",test);

  test = myx_t(t, "ResourceStrings", "FetchUsernames", "son scheiss..");
  printf("%s\n",test);

  test = myx_tg(t, "ResourceStrings", "PingBtn", "n/a");
  printf("%s\n",test);

  test = myx_tg(t, "ResourceStrings", "DetailsBtn_expanded", "n/a");
  printf("%s\n",test);

  myx_free_trans(t);

  return 0;
}*/

/*returns NULL on error, for details check error_code then*/
MYX_TRANS* myx_init_trans(const char* translation_file_name, const char* application_translation_file_name, const char *language, MYX_LIB_ERROR *error_code)
{
  MYX_TRANS *trans;

  *error_code= MYX_NO_ERROR;

  trans = calloc(1, sizeof(MYX_TRANS) );

  if (parse_translation_file(trans, translation_file_name, language, error_code, 2) )
  {
    free(trans);
    return NULL;
  }

  if (parse_translation_file(trans, application_translation_file_name,language, error_code, 1) )
  {
    free(trans);
    return NULL;
  }

  sort_alphabetically(trans);

  return trans;
}


int myx_free_trans(MYX_TRANS *trans)
{
  unsigned int i;

  if (trans)
  {
    for (i=0; i< trans->textgroups_general_num; i++)
    {
      free_textgroup_content(trans->textgroups_general+i);
    }
    free(trans->textgroups_general);

    for (i=0; i< trans->textgroups_num; i++)
    {
      free_textgroup_content(trans->textgroups+i);
    }
    free(trans->textgroups);
    free(trans);
  }

  return 0;
}


/*gets a translation of a string with id_text/id_group from the application-specific translation file*/
/*returns the orginial_text if not found*/
/*generally the caller must not free the return value. (he may free if the return value=original_text*/
const char * myx_t(MYX_TRANS *trans, const char *id_textgroup, const char *id_text, const char *original_text)
{
  MYX_TRANS_TEXTGROUP *tg;
  const char *result = original_text;
  MYX_TRANS_TEXT *text;
  MYX_TRANS_TEXTGROUP tmp_tg = {(char *)id_textgroup,0,NULL};
  MYX_TRANS_TEXT tmp_text = {(char *)id_text, NULL};

  if(trans)
  {
    /*get a pointer to the textgroup with id 'id_textgroup'*/
    tg= NULL;
    if (trans->current_textgroup != NULL)
    {
      if (!xmlStrcmp(trans->current_textgroup->id, id_textgroup) )
      {
        tg= trans->current_textgroup;
      }
    }

    if(!tg)
    {
      tg= (MYX_TRANS_TEXTGROUP*) bsearch(&tmp_tg,trans->textgroups, trans->textgroups_num, sizeof(MYX_TRANS_TEXTGROUP), compare_textgroups);
      if (tg == NULL) return result;
      trans->current_textgroup = tg;
    }

    /*now find the text with id_text in the group*/
    text = (MYX_TRANS_TEXT *) bsearch(&tmp_text,tg->texts,tg->texts_num, sizeof(MYX_TRANS_TEXT), compare_transtexts);
    if (text != NULL) result = text->trans;

    return result;
  }
  else
  {
    return original_text;
  }
}

/*gets a translation of a string with id_text/id_group from the general translation file*/
/*returns the orginial_text if not found*/
/*generally the caller must not free the return value. (he may free if the return value=original_text*/
const char * myx_tg(MYX_TRANS *trans, const char *id_textgroup, const char *id_text, const char *original_text)
{
  MYX_TRANS_TEXTGROUP *tg;
  const char *result = original_text;
  MYX_TRANS_TEXT *text;
  MYX_TRANS_TEXTGROUP tmp_tg = {(char *)id_textgroup,0,NULL};
  MYX_TRANS_TEXT tmp_text = {(char *)id_text, NULL};

  /*get a pointer to the textgroup with id 'id_textgroup'*/
  if (trans->current_textgroup_general != NULL && !xmlStrcmp(trans->current_textgroup_general->id, id_textgroup) )
  {
    tg = trans->current_textgroup_general;
  }
  else
  {
    tg = (MYX_TRANS_TEXTGROUP*) bsearch(&tmp_tg,trans->textgroups_general, trans->textgroups_general_num, sizeof(MYX_TRANS_TEXTGROUP), compare_textgroups);
    if (tg == NULL) return result;
    trans->current_textgroup_general = tg;
  }

  /*now find the text with id_text in the group*/
  text = (MYX_TRANS_TEXT *) bsearch(&tmp_text,tg->texts,tg->texts_num, sizeof(MYX_TRANS_TEXT), compare_transtexts);
  if (text != NULL) result = text->trans;

  return result;
}





/**********************/
/*   PRIVATE FUNCTIONS*/
/**********************/


static int compare_transtexts(const void *a, const void *b)
{
  MYX_TRANS_TEXT *t1 = (MYX_TRANS_TEXT *) a;
  MYX_TRANS_TEXT *t2 = (MYX_TRANS_TEXT *) b;

  return ( xmlStrcmp(t1->id, t2->id) );
}

static int compare_textgroups(const void *a, const void *b)
{
  MYX_TRANS_TEXTGROUP *g1 = (MYX_TRANS_TEXTGROUP *) a;
  MYX_TRANS_TEXTGROUP *g2 = (MYX_TRANS_TEXTGROUP *) b;

  return ( xmlStrcmp(g1->id,g2->id) );
}

static void sort_alphabetically(MYX_TRANS *trans)
{
  unsigned int i;

  /*sort the groups*/
  if (trans->textgroups)
    qsort(trans->textgroups, trans->textgroups_num, sizeof(MYX_TRANS_TEXTGROUP), compare_textgroups);

  if (trans->textgroups_general)
    qsort(trans->textgroups_general, trans->textgroups_general_num, sizeof(MYX_TRANS_TEXTGROUP), compare_textgroups);


  /*sort the text-elements in their groups*/
  for (i=0; i < trans->textgroups_num; i++)
  {
    if (trans->textgroups[i].texts)
      qsort(trans->textgroups[i].texts, trans->textgroups[i].texts_num, sizeof(MYX_TRANS_TEXT), compare_transtexts);
  }

  for (i=0; i < trans->textgroups_general_num; i++)
  {
    if (trans->textgroups_general[i].texts)
      qsort(trans->textgroups_general[i].texts, trans->textgroups_general[i].texts_num, sizeof(MYX_TRANS_TEXT), compare_transtexts);
  }
}

static int parse_translation_file(MYX_TRANS *trans, const char *filename,const char *lang, MYX_LIB_ERROR *error_code,  FILE_KIND kind)
{
  xmlDocPtr doc;
  xmlNodePtr root, cur; 
  int i;

  if (!file_exists(filename))
  {
    *error_code = MYX_ERROR_CANT_OPEN_FILE;
    return -1;
  }

  doc = xmlParseFile(filename);
    
  if (doc == NULL ) {
    *error_code = MYX_XML_PARSE_ERROR;
    return -1;
  }
    
  root = xmlDocGetRootElement(doc);
    
  if (root == NULL) {
    *error_code = MYX_XML_EMPTY_DOCUMENT;
    xmlFreeDoc(doc);
    return -1;
  }
    
  if (xmlStrcmp(root->name, (const xmlChar *) "messages")) 
  {
    *error_code =  MYX_XML_NO_VALID_DOCUMENT;
    xmlFreeDoc(doc);
    return -1;
  }

  if (kind == APPLICATION_SPECIFIC_FILE)
  {
    trans->textgroups_num = get_child_count(root, "textgroup");
    if (trans->textgroups_num > 0)
      trans->textgroups = calloc(trans->textgroups_num, sizeof(MYX_TRANS_TEXTGROUP));
  }
  else if (kind == GLOBAL_FILE)
  {
    trans->textgroups_general_num = get_child_count(root, "textgroup");
    if (trans->textgroups_general_num > 0)
      trans->textgroups_general = calloc(trans->textgroups_general_num, sizeof(MYX_TRANS_TEXTGROUP));
  }
  else assert(0);

  i=0;
  cur = root->children;
  while (cur)
  {
    if (!xmlStrcmp(cur->name, "textgroup") )
    {
      if (kind == APPLICATION_SPECIFIC_FILE)
	read_in_textgroup(cur, trans->textgroups+i, lang);
      else
	read_in_textgroup(cur, trans->textgroups_general+i, lang);

      i++;
    }

    cur = cur->next;
  }


  return 0;
}


static void read_in_textgroup(const xmlNodePtr textgroup_node, MYX_TRANS_TEXTGROUP *group, const char *lang)
{
  int i;
  xmlNodePtr cur;

  group->id = xmlGetProp(textgroup_node, "id");

  group->texts_num = get_child_count(textgroup_node, "text");
  if (group->texts_num > 0)
    group->texts = calloc(group->texts_num, sizeof(MYX_TRANS_TEXT) );


  i = 0;
  cur = textgroup_node->children;
  while (cur)
  {
    if ( !xmlStrcmp(cur->name, "text") &&  cur->type == XML_ELEMENT_NODE)
    {
      read_in_text(cur, group->texts+i, lang);
      i++;
    }

    cur = cur->next;
  }

}

static void read_in_text(const xmlNodePtr text_node, MYX_TRANS_TEXT *text, const char *lang)
{
  xmlDocPtr doc = text_node->doc;
  xmlNodePtr cur;
  xmlChar *tmp;

  text->id = xmlGetProp(text_node, "id");

  cur = text_node->children;
  while (cur)
  {
    if (! xmlStrcmp(cur->name, "trans") )
    {
      tmp = xmlGetProp(cur, "lang");
      if (! xmlStrcmp(tmp, lang) )
      {
	text->trans = xmlNodeListGetString(doc,cur->children,1);
	xmlFree(tmp);
	break;
      }
      xmlFree(tmp);
    }
    cur = cur->next;
  }
}



static void free_textgroup_content(MYX_TRANS_TEXTGROUP *group)
{
  unsigned int i;

  xmlFree(group->id);

  for(i=0; i< group->texts_num; i++)
  {
    free_text_content(group->texts+i);
  }
  free(group->texts);
}

static void free_text_content(MYX_TRANS_TEXT *text)
{
  xmlFree(text->id);
  xmlFree(text->trans);
}
