/* Copyright (C) 2004 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include "MGPreferencesEditor.h"

#define _(s) s

#define DEFAULT_MYSQL_PORT 3306





MGPreferencesEditor::MGPreferencesEditor(GtkWindow *widget)
  : Gtk::Window(widget), _connections(0)
{
}


MGPreferencesEditor::~MGPreferencesEditor()
{
  delete _xml;
  delete _connections;
}


MGPreferencesEditor *MGPreferencesEditor::create(MGPreferences *prefs)
{
  MGGladeXML *xml= new MGGladeXML(APP_DATADIR"/preferences.glade");

  MGPreferencesEditor *editor= 0;
  xml->get_widget_derived("preferences_window", editor);
  editor->_xml= xml;
  editor->_prefs= prefs;

  editor->init();
  
  return editor;
}


void MGPreferencesEditor::init()
{
  Gtk::Widget *heading= get_widget("heading");
  heading->signal_expose_event().connect(SigC::slot(*this,&MGPreferencesEditor::expose_heading));

  ((Gtk::ToggleButton*)get_widget("administrator_radio"))->
    signal_toggled().connect(SigC::bind<int>(SigC::slot(*this,&MGPreferencesEditor::menu_changed),0));

  ((Gtk::ToggleButton*)get_widget("general_radio"))->
    signal_toggled().connect(SigC::bind<int>(SigC::slot(*this,&MGPreferencesEditor::menu_changed),1));

  ((Gtk::ToggleButton*)get_widget("connections_radio"))->
    signal_toggled().connect(SigC::bind<int>(SigC::slot(*this,&MGPreferencesEditor::menu_changed),2));

  ((Gtk::Button*)get_widget("close_button"))->signal_clicked().connect(SigC::slot(*this,&MGPreferencesEditor::close_clicked));
  ((Gtk::Button*)get_widget("discard_button"))->signal_clicked().connect(SigC::slot(*this,&MGPreferencesEditor::discard_clicked));
  ((Gtk::Button*)get_widget("apply_button"))->signal_clicked().connect(SigC::slot(*this,&MGPreferencesEditor::apply_clicked));
  
  signal_delete_event().connect(SigC::slot(*this,&MGPreferencesEditor::close_window));
  
  // init subpanels
  init_general();
  init_connections();
  init_administrator();

  _dirty= false;
  update_sensitivity();
  
  menu_changed(0);
}


void MGPreferencesEditor::show()
{
  Gtk::Window::show();
  
  update();
}


void MGPreferencesEditor::update()
{
  MYX_LIB_ERROR merror;
  MYX_USER_CONNECTIONS *conns;

  conns= myx_load_user_connections(_prefs->build_path_to(_prefs->connections_filename).c_str(),
                                          &merror);
  if (conns)
  {
    if (_connections)
      delete _connections;
    _connections= new UserConnectionList(conns);
    myx_free_user_connections(conns);
  }
  else
  {
    _connections= new UserConnectionList;
  }
  
  update_administrator();
  update_general();
  update_connections();
}


bool MGPreferencesEditor::close_window(GdkEventAny *ev)
{
  close_clicked();

  return true;
}


void MGPreferencesEditor::discard_clicked()
{
  update_general();
  update_connections();
  update_administrator();
  unset_dirty();
}


void MGPreferencesEditor::apply_clicked()
{
  commit_general();
  commit_connections();
  commit_administrator();
  
  _prefs->save();
  
  unset_dirty();
}


void MGPreferencesEditor::close_clicked()
{
  hide();
}


bool MGPreferencesEditor::expose_heading(GdkEventExpose *expose)
{
  if (expose->count == 0)
  {
    Gtk::Widget *heading= get_widget("heading");
    Glib::RefPtr<Gdk::Window> win= heading->get_window();
    Glib::RefPtr<Gtk::Style> style= heading->get_style();
    int w= heading->get_width();
    int h= heading->get_height();

    win->clear();
    win->draw_rectangle(style->get_dark_gc(Gtk::STATE_NORMAL), true,
                        0, 0, w, h);
    win->draw_line(style->get_black_gc(), 0, 0, w, 0);
    win->draw_line(style->get_black_gc(), 0, 0, 0, h);

    win->draw_line(style->get_white_gc(), 0, h-1, w, h-1);
    win->draw_line(style->get_white_gc(), w-1, 0, w-1, h);
  
    Glib::RefPtr<Pango::Layout> l= create_pango_layout("");
    int tw, th;
    l->set_markup("<b>"+_heading+"</b>");
    l->get_pixel_size(tw, th);
    win->draw_layout(style->get_white_gc(), 5, (h - th) / 2, l);
  }
  return true;
}


void MGPreferencesEditor::menu_changed(int page)
{
  switch (page)
  {
  case 0:
    _heading= _("MySQL Administrator Options");
    break;
  case 1:
    _heading= _("General Options");
    break;
  case 2:
    _heading= _("Connections");
    break;
  }

  get_widget("heading")->queue_draw();

  ((Gtk::Notebook*)get_widget("notebook"))->set_current_page(page);
}


void MGPreferencesEditor::update_sensitivity()
{
  get_widget("apply_button")->set_sensitive(_dirty);
  get_widget("discard_button")->set_sensitive(_dirty);
}


void MGPreferencesEditor::set_dirty()
{
  if (_dirty != true)
  {
    _dirty= true;
    update_sensitivity();
  }
}

void MGPreferencesEditor::unset_dirty()
{
  if (_dirty != false)
  {
    _dirty= false;
    update_sensitivity();
  }
}


//------------------------------------------------------------

void MGPreferencesEditor::init_administrator()
{
  ((Gtk::ToggleButton*)get_widget("admin_show_global_check"))->
    signal_toggled().connect(SigC::slot(*this,&MGPreferencesEditor::set_dirty));

  ((Gtk::ToggleButton*)get_widget("admin_show_table_check"))->
    signal_toggled().connect(SigC::slot(*this,&MGPreferencesEditor::set_dirty));
}


void MGPreferencesEditor::update_administrator()
{
  ((Gtk::ToggleButton*)get_widget("admin_show_global_check"))->set_active(_prefs->show_user_global_privileges);
  ((Gtk::ToggleButton*)get_widget("admin_show_table_check"))->set_active(_prefs->show_user_table_column_privileges);
}


void MGPreferencesEditor::commit_administrator()
{
  _prefs->show_user_global_privileges= ((Gtk::ToggleButton*)get_widget("admin_show_global_check"))->get_active();
  _prefs->show_user_table_column_privileges= ((Gtk::ToggleButton*)get_widget("admin_show_table_check"))->get_active();
}


//------------------------------------------------------------

void MGPreferencesEditor::init_general()
{
  Gtk::CheckButton *ck= (Gtk::CheckButton*)get_widget("general_store_pw_check");

  ck->signal_clicked().connect(SigC::slot(*this,&MGPreferencesEditor::general_pw_store_clicked));
  ck->signal_toggled().connect(SigC::slot(*this,&MGPreferencesEditor::set_dirty));

  ((Gtk::OptionMenu*)get_widget("general_store_pw_menu"))->signal_changed().connect(SigC::slot(*this,&MGPreferencesEditor::set_dirty));
}


void MGPreferencesEditor::general_pw_store_clicked()
{
  Gtk::CheckButton *ck= (Gtk::CheckButton*)get_widget("general_store_pw_check");
  
  get_widget("general_store_pw_box")->set_sensitive(ck->get_active());
}


void MGPreferencesEditor::update_general()
{
  if (_prefs->password_storage_type == MYX_PASSWORD_NOT_STORED)
  {
    ((Gtk::ToggleButton*)get_widget("general_store_pw_check"))->set_active(false);
  }
  else
  {
    ((Gtk::ToggleButton*)get_widget("general_store_pw_check"))->set_active(true);
    ((Gtk::OptionMenu*)get_widget("general_store_pw_menu"))->set_history((int)_prefs->password_storage_type-(int)MYX_PASSWORD_OBSCURED);
  }
  general_pw_store_clicked();
}


void MGPreferencesEditor::commit_general()
{
  if (((Gtk::ToggleButton*)get_widget("general_store_pw_check"))->get_active())
    _prefs->password_storage_type= (MYX_PASSWORD_STORAGE_TYPE)(((Gtk::OptionMenu*)get_widget("general_store_pw_menu"))->get_history()+(int)MYX_PASSWORD_PLAINTEXT);
  else
    _prefs->password_storage_type= MYX_PASSWORD_NOT_STORED;
}

//------------------------------------------------------------


static MGAdvancedOption option_list[]=
{
  {"COMPRESS", "option1_check", 'B'},
  {"USE_SSL", "option2_check", 'B'},
  {"SOCKET_PATH", "option1_entry", 'S'},
  {NULL, NULL, 0}
};


void MGPreferencesEditor::init_connections()
{
#define CONNECT_CHANGED(type,name) ((type*)get_widget(name))->signal_changed().connect(SigC::bind<const char*>(SigC::slot(*this,&MGPreferencesEditor::connection_data_changed),name))
  CONNECT_CHANGED(Gtk::Entry, "connection_entry");
  CONNECT_CHANGED(Gtk::Entry, "username_entry");
  CONNECT_CHANGED(Gtk::Entry, "password_entry");
  CONNECT_CHANGED(Gtk::Entry, "hostname_entry");
  CONNECT_CHANGED(Gtk::SpinButton, "port_spin");
  CONNECT_CHANGED(Gtk::Entry, "schema_entry");
  ((Gtk::TextView*)get_widget("connection_note_text"))->get_buffer()->signal_changed().connect(SigC::bind<const char*>(SigC::slot(*this,&MGPreferencesEditor::connection_data_changed),"connection_note_text"));

  ((Gtk::Button*)get_widget("connection_add_button"))->signal_clicked().connect(SigC::slot(*this,&MGPreferencesEditor::connection_add));
  ((Gtk::Button*)get_widget("connection_remove_button"))->signal_clicked().connect(SigC::slot(*this,&MGPreferencesEditor::connection_remove));
  
  for (unsigned int i= 0; option_list[i].name; i++)
  {
    switch (option_list[i].type)
    {
    case 'B':
      ((Gtk::ToggleButton*)get_widget(option_list[i].widget))->signal_toggled().connect(SigC::bind<const char*>(SigC::slot(*this,&MGPreferencesEditor::connection_data_changed),option_list[i].widget));
      break;
    case 'S':
      ((Gtk::Entry*)get_widget(option_list[i].widget))->signal_changed().connect(SigC::bind<const char*>(SigC::slot(*this,&MGPreferencesEditor::connection_data_changed),option_list[i].widget));
      break;
    }
  }
  
#undef CONNECT_CHANGED

  Gtk::TreeView *tree= (Gtk::TreeView*)get_widget("connection_tree");

  tree->get_selection()->set_mode(Gtk::SELECTION_BROWSE);
  tree->get_selection()->signal_changed().connect(SigC::slot(*this,&MGPreferencesEditor::connection_show));

  tree->append_column("", _connection_columns._icon);
  tree->append_column("", _connection_columns._text);

  _connection_store= Gtk::TreeStore::create(_connection_columns);

  tree->set_model(_connection_store);

  _conn_group_icon= Gdk::Pixbuf::create_from_file(APP_DATADIR"/folder_16x16.png");
  _conn_icon= Gdk::Pixbuf::create_from_file(APP_DATADIR"/networkhost_16x16.png");
  
  _connection_ignore_changes= false;
}


void MGPreferencesEditor::update_connections()
{
  Gtk::TreeIter conn_iter, hist_iter;
  Gtk::TreeModel::Row row;

  _connection_store->clear();

  conn_iter= _connection_store->append();
  row= *conn_iter;
  row[_connection_columns._icon]= _conn_group_icon;
  row[_connection_columns._text]= _("Connections");
  row[_connection_columns._is_sep]= true;

  hist_iter= _connection_store->append();
  row= *hist_iter;
  row[_connection_columns._icon]= _conn_group_icon;
  row[_connection_columns._text]= _("History");
  row[_connection_columns._is_sep]= true;

  for (std::vector<UserConnection>::const_iterator conn= _connections->connections.begin();
       conn != _connections->connections.end(); ++conn)
  {
    Gtk::TreeIter iter;

    if (conn->storage_type==MYX_FAVORITE_USER_CONNECTION)
      iter= _connection_store->append((*conn_iter).children());
    else
      iter= _connection_store->append((*hist_iter).children());

    row= *iter;
    row[_connection_columns._icon]= _conn_icon;
    row[_connection_columns._text]= conn->connection_name.empty() ?"history":conn->connection_name;
    row[_connection_columns._is_sep]= false;
    row[_connection_columns._data]= *conn;
  }

  ((Gtk::TreeView*)get_widget("connection_tree"))->expand_all();

  get_widget("connection_notebook")->set_sensitive(false);
}


void MGPreferencesEditor::commit_connections()
{
  _connections->connections.clear();
  _connections->last_connection= -1;

  // go through everything adding the connections in the list

  Gtk::TreeIter sel= ((Gtk::TreeView*)get_widget("connection_tree"))->get_selection()->get_selected();
  int idx= 0;
  Gtk::TreeIter iter, piter;
  Gtk::TreeModel::Row row, prow;
  for (piter= _connection_store->children().begin(); 
       piter != _connection_store->children().end(); ++piter, ++idx)
  {
    prow= *piter;
    for (iter= prow.children().begin(); iter != prow.children().end(); ++iter)
    {
      row= *iter;
      UserConnection user(row[_connection_columns._data]);
      
      _connections->connections.push_back(user);
      if (sel == iter)
        _connections->last_connection= idx;
    }
  }

  // save
  myx_store_user_connections(_connections->get_obj(),
                             _prefs->password_storage_type,
                             _prefs->build_path_to(_prefs->connections_filename).c_str());
}


void MGPreferencesEditor::connection_add()
{
  UserConnection con;

  con.connection_name= "new connection";
  con.username= "";
  con.password= "";
  con.hostname= "";
  con.port= DEFAULT_MYSQL_PORT;
  con.schema= "";
  con.connection_type= MYX_MYSQL_CONN;
  con.storage_type= MYX_FAVORITE_USER_CONNECTION;

  Gtk::TreeModel::Row prow= *_connection_store->children().begin();
  
  Gtk::TreeView *tree= (Gtk::TreeView*)get_widget("connection_tree");
  Gtk::TreeIter iter= _connection_store->append(prow.children());
  Gtk::TreeModel::Row row= *iter;

  tree->expand_to_path(Gtk::TreePath(_connection_store->children().begin()));

  row[_connection_columns._icon]= _conn_icon;
  row[_connection_columns._text]= con.connection_name;
  row[_connection_columns._is_sep]= false;
  row[_connection_columns._data]= con;

  tree->get_selection()->select(iter);
  
  set_dirty();
}


void MGPreferencesEditor::connection_remove()
{
  Gtk::TreeView *tree= (Gtk::TreeView*)get_widget("connection_tree");
  Gtk::TreeIter iter= tree->get_selection()->get_selected();

  _connection_store->erase(iter);
  
  set_dirty();
}


void MGPreferencesEditor::connection_show()
{
  Gtk::TreeView *tree= (Gtk::TreeView*)get_widget("connection_tree");
  Gtk::TreeIter iter= tree->get_selection()->get_selected();
  UserConnection user;
  bool sens= false;

  _connection_ignore_changes= true;
  
  if (iter)
  {
    Gtk::TreeModel::Row row= *iter;

    if (!row[_connection_columns._is_sep])
    {
      user= row[_connection_columns._data];
      sens= true;
      
      get_widget("connection_remove_button")->set_sensitive(true);
    }
    else
    {
      get_widget("connection_remove_button")->set_sensitive(false);
    }
  }

  ((Gtk::Entry*)get_widget("connection_entry"))->set_text(user.connection_name);
  ((Gtk::Entry*)get_widget("username_entry"))->set_text(user.username);
  ((Gtk::Entry*)get_widget("password_entry"))->set_text(user.password);
  ((Gtk::Entry*)get_widget("hostname_entry"))->set_text(user.hostname);
  ((Gtk::SpinButton*)get_widget("port_spin"))->set_value(user.port);
  ((Gtk::Entry*)get_widget("schema_entry"))->set_text(user.schema);
  
  ((Gtk::TextView*)get_widget("connection_note_text"))->get_buffer()->set_text(user.notes);

  myg_show_advanced_options(_xml, option_list, user.advanced_options);

  get_widget("connection_notebook")->set_sensitive(sens);

  _connection_ignore_changes= false;
}


void MGPreferencesEditor::connection_data_changed(const char *field)
{
  Gtk::TreeIter iter= ((Gtk::TreeView*)get_widget("connection_tree"))->get_selection()->get_selected();

  if (_connection_ignore_changes || !iter)
    return;
    
  Gtk::TreeModel::Row row= *iter;

  if (strcmp(field, "connection_entry")==0)
  {    
    row[_connection_columns._text]= ((Gtk::Entry*)get_widget(field))->get_text();
  }

  UserConnection user= row[_connection_columns._data];

  user.connection_name= ((Gtk::Entry*)get_widget("connection_entry"))->get_text();
  user.username= ((Gtk::Entry*)get_widget("username_entry"))->get_text();
  user.password= ((Gtk::Entry*)get_widget("password_entry"))->get_text();
  user.hostname= ((Gtk::Entry*)get_widget("hostname_entry"))->get_text();
  user.port= ((Gtk::SpinButton*)get_widget("port_spin"))->get_value_as_int();
  user.schema= ((Gtk::Entry*)get_widget("schema_entry"))->get_text();
  user.notes= ((Gtk::TextView*)get_widget("connection_note_text"))->get_buffer()->get_text();

  myg_fetch_advanced_options(_xml, option_list, user.advanced_options);

  row[_connection_columns._data]= user;

  set_dirty();
}


//bool MGPreferencesEditor::connection_validate()
