/* Copyright (C) 2003 MySQL AB

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/*
 * Generic database server connection dialog for MySQL for Gtk-2
 * This requires the counterpart glade 2 file.
 * 
 * MAIN ENTRY POINTS
 *     myg_connect_with_dialog()
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <glade/glade.h>

#include <mysql.h>

#include <pthread.h>

#include "myx_aux_functions.h"
#include "myx_public_interface.h"

#include "myg_gtk_utils.h"

#define _(a) a


#define ZSTRDUP(s) ((s&&!is_str_blank(s))?(char*)str_trim(strdup(s)):NULL)

#define MYG_FLAG_PICK_SCHEMA   (1<<0)



#define PROGRESS_INDICATOR_FILE "progress_indicator.gif"

#define GLADE_FILE "connect_dialog.glade"


typedef struct {
  GladeXML *xml;

  MYX_USER_CONNECTIONS *connections;

  MYX_USER_CONNECTION *cur_conn;
  
  const char *app_title;
  
  MYSQL *mysql;

  int conn_result;
  pthread_t conn_thread;

  GtkWidget *ping_view;
  FILE *ping_output;
  GtkDialog *ping_dialog;

  unsigned int advanced_shown:1;
  unsigned int pick_schema:1;
  unsigned int connecting:1;

} myg_connect_dialog;



typedef struct {
  char *name;
  GtkSignalFunc func;
} glade_handler;



static void start_connection(myg_connect_dialog *dlg);
static void stop_connection(myg_connect_dialog *dlg);
static void update_connect_sensitivity(myg_connect_dialog *dlg);

static void load_connection_list(myg_connect_dialog *dlg, const char *clpath);

static void reset_connection_values(myg_connect_dialog *dlg);
static void display_connection_values(myg_connect_dialog *dlg,
                                      MYX_USER_CONNECTION *conn);
static MYX_USER_CONNECTION *fetch_connection_values(myg_connect_dialog *dlg,
                                                    MYX_USER_CONNECTION *conn);

static const char *disable_while_connecting[]=
{
  "connection_list",
      "hostname_entry",
      "password_entry",
      "username_entry",
      "port_spin",
      "schema_entry",
      "advanced_frame",
      "details_button",
      "clear_button",
      "connect_button",
      NULL
};


/* signal handlers */

static void on_connect_button_clicked(GtkWidget *w, gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;
  
  start_connection(dlg);
}


static void on_clear_button_clicked(GtkWidget *w, gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;

  reset_connection_values(dlg);
}


static void on_cancel_button_clicked(GtkWidget *w, gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;

  if (dlg->conn_thread)
      stop_connection(dlg);
  else
      gtk_main_quit();
}


static int find_menu_index_of_current_connection(myg_connect_dialog *dlg)
{
  GtkWidget *opm= glade_xml_get_widget(dlg->xml, "connection_list");
  GtkWidget *menu= glade_xml_get_widget(dlg->xml, "connection_menu");
  MYX_USER_CONNECTION *conn;
  GtkWidget *wid;
  GList *list= GTK_MENU_SHELL(menu)->children;
  const char *username;
  const char *hostname;
  const char *schema;
  int index;
  
  wid= glade_xml_get_widget(dlg->xml, "username_entry");
  username= gtk_entry_get_text(GTK_ENTRY(wid));

  wid= glade_xml_get_widget(dlg->xml, "hostname_entry");
  hostname= gtk_entry_get_text(GTK_ENTRY(wid));

  wid= glade_xml_get_widget(dlg->xml, "schema_entry");
  schema= gtk_entry_get_text(GTK_ENTRY(wid));

  /* locate menu item for the typed in stuff */
  for (list= g_list_first(list), index= 0; list != NULL; 
       list= g_list_next(list), index++)
  {
    GtkWidget *item= (GtkWidget*)list->data;

    conn= (MYX_USER_CONNECTION*)gtk_object_get_data(GTK_OBJECT(item),
                                                    "user_data");
    if (conn 
        && strcmp(conn->username, username)==0
        && strcmp(conn->hostname, hostname)==0
        && (!conn->schema || strcmp(conn->schema, schema)==0))
    {
      return index;
    }
  }

  return -1;
}


static void on_check_entry(GtkWidget *w, gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;

  update_connect_sensitivity(dlg);
}


static void on_details_button_clicked(GtkWidget *w, gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;
  GtkWidget *frame, *btn;

  frame= glade_xml_get_widget(dlg->xml, "advanced_frame");
  btn= glade_xml_get_widget(dlg->xml, "details_button");
  
  if (dlg->advanced_shown)
  {
    dlg->advanced_shown= 0;
    gtk_widget_hide(frame);
    
    gtk_button_set_label(GTK_BUTTON(btn), _("Details >>"));
  }
  else
  {
    dlg->advanced_shown= 1;
    gtk_widget_show(frame);

    gtk_button_set_label(GTK_BUTTON(btn), _("Details <<"));
  }
}


static void on_connection_list_changed(GtkWidget *w, gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;
  MYX_USER_CONNECTION *conn;
  GtkWidget *item= GTK_OPTION_MENU(w)->menu_item;

  conn= gtk_object_get_data(GTK_OBJECT(item), "user_data");
  
  if (conn)
  {
    display_connection_values(dlg, conn);
  }
  
  dlg->cur_conn= conn;
}



/* --- */

static glade_handler glade_handler_list[]=
{
  { "on_connect_button_clicked", (GtkSignalFunc)on_connect_button_clicked },
  { "on_clear_button_clicked", (GtkSignalFunc)on_clear_button_clicked },
  { "on_cancel_button_clicked", (GtkSignalFunc)on_cancel_button_clicked },
  { "on_entry_changed", (GtkSignalFunc)on_check_entry },
  { "on_details_button_clicked", (GtkSignalFunc)on_details_button_clicked },
  { "on_connection_list_changed", (GtkSignalFunc)on_connection_list_changed },
};


myg_connect_dialog *myg_init_connect_dialog(const char *clpath,
                                            int flags)
{
  myg_connect_dialog *dlg= g_malloc(sizeof(myg_connect_dialog));
  GtkWidget *prog;
  GdkPixbufAnimation *anim;
  GError *error= NULL;
  GtkWidget *wid;
  int i;

  memset(dlg, 0, sizeof(myg_connect_dialog));
  
  dlg->xml= glade_xml_new(APP_DATADIR"/"GLADE_FILE, NULL, NULL);
  if (!dlg->xml)
  {
    g_free(dlg);
    return NULL;
  }

  dlg->pick_schema= (flags&MYG_FLAG_PICK_SCHEMA)!=0;


  /* enable history in text entries */

  wid= glade_xml_get_widget(dlg->xml, "username_entry");
  myg_entry_history_attach(wid);

  wid= glade_xml_get_widget(dlg->xml, "hostname_entry");
  myg_entry_history_attach(wid);

  wid= glade_xml_get_widget(dlg->xml, "schema_entry");
  myg_entry_history_attach(wid);

  /* connect signals */

  for (i= 0; i < sizeof(glade_handler_list)/sizeof(glade_handler); i++) 
  {
    glade_xml_signal_connect_data(dlg->xml, glade_handler_list[i].name,
                                  glade_handler_list[i].func,
                                  dlg);
  }

  /* setup progress animation thingy (libglade seems to be buggy, it loads
   the image as a static image) */
  prog= glade_xml_get_widget(dlg->xml, "progress_image");

  anim= gdk_pixbuf_animation_new_from_file(APP_DATADIR"/"PROGRESS_INDICATOR_FILE, &error);
  if (!anim) 
  {
    if (error)
    {
      printf("%s\n",error->message);
    }
    g_error_free(error);
  }
  else
  {
    gtk_image_set_from_animation(GTK_IMAGE(prog), anim);
  }
  
  if (dlg->pick_schema)
  {
    gtk_widget_show(glade_xml_get_widget(dlg->xml, "schema_label"));
    gtk_widget_show(glade_xml_get_widget(dlg->xml, "schema_entry"));
  }
  else
  {
    gtk_widget_hide(glade_xml_get_widget(dlg->xml, "schema_label"));
    gtk_widget_hide(glade_xml_get_widget(dlg->xml, "schema_entry"));
  }

  load_connection_list(dlg, clpath);

  update_connect_sensitivity(dlg);
  
  return dlg;
}



/*
 *----------------------------------------------------------------------
 *
 * myg_connect_with_dialog - shows mysql connection dialog and connect
 *
 * SYNOPSIS
 * 
 *   Shows a dialog requesting connection parameters and attempt to 
 *   connect to the server. The dialog GUI is read from the glade file
 *   in APP_DATADIR/GLADE_FILE
 * 
 * DESCRIPTION
 *
 * RETURN VALUE
 * 
 *   A mysql connection object or NULL on error or cancel by user.
 *
 * NOTES
 *
 *   Uses data stored in clpath (user connection parameters).
 * 
 *----------------------------------------------------------------------
 */
MYSQL *myg_connect_with_dialog(const char *clpath,
                               const char *title,
                               int flags,
                               MYX_USER_CONNECTION **conn_ret)
{
  myg_connect_dialog *dlg= 0;
  MYSQL *mysql;
  GtkWidget *win;
  
  if (!dlg) 
  {
    dlg= myg_init_connect_dialog(clpath, flags);
    if (!dlg)
    {
      GtkWidget *mbox;
    
      mbox= gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL,
                                   GTK_MESSAGE_ERROR,
                                   GTK_BUTTONS_OK,
                                   _("Can't show connection dialog.\n"
                                     "Software installation possibly corrupted or incorrect."));
      gtk_dialog_run(GTK_DIALOG(mbox));
      gtk_widget_destroy(mbox);

      return NULL;
    }
  }

  win= glade_xml_get_widget(dlg->xml, "connect_dialog");
    
  gtk_window_set_title(GTK_WINDOW(win), title);

  gtk_widget_show(win);

  dlg->app_title= title;
  
  gtk_main();

  gtk_widget_hide(win);

  mysql= dlg->mysql;

  if (mysql) /* successful connection */
  {
    MYX_USER_CONNECTION *conn;
    int i;

    /* store user connection if not in list */
    for (i= 0; i < dlg->connections->user_connections_num; i++)
    {
      conn= dlg->connections->user_connections+i;

      if (conn == dlg->cur_conn)
          break;
    }

    if (i == dlg->connections->user_connections_num)
    {
      dlg->connections= realloc(dlg->connections->user_connections,
                                dlg->connections->user_connections_num
                                * sizeof(MYX_USER_CONNECTION));
      memcpy(dlg->connections+i, dlg->cur_conn, sizeof(MYX_USER_CONNECTION));
    }

    /* store last connection */
    dlg->connections->last_connection= i;

    if (clpath)
    {
      /* save connection list */
      myx_store_user_connections(dlg->connections,
                                 MYX_PASSWORD_NOT_STORED,
                                 clpath);
    }
    
    *conn_ret= dlg->cur_conn;
  }

  /* cleanup */
  dlg->mysql= NULL;

  
  return mysql;
}



static int is_str_blank(const char *s)
{
  while (*s) 
  {
    if (!isspace(*s))
        return 0;
    s++;
  }
  return 1;
}


static void load_connection_list(myg_connect_dialog *dlg, const char *clpath)
{
  MYX_USER_CONNECTIONS *connection_list;
  MYX_LIB_ERROR merror;
  GtkWidget *username_e, *hostname_e, *schema_e;
  int i;

  username_e= glade_xml_get_widget(dlg->xml, "username_entry");      
  hostname_e= glade_xml_get_widget(dlg->xml, "hostname_entry");      
  schema_e= glade_xml_get_widget(dlg->xml, "schema_entry");

  /* load connection list data */
  if (clpath)
      connection_list= myx_load_user_connections(clpath, &merror);
  else
      connection_list= NULL;
  if (connection_list && merror == MYX_NO_ERROR)
  {
    GtkWidget *wid;
    GtkWidget *menu;
    GtkWidget *item;
    MYX_USER_CONNECTION *conn;
    char *tmp;
    int unnamed= 1;
    int len= strlen(_("* Unnamed %i *"))+1;

    menu= glade_xml_get_widget(dlg->xml, "connection_menu");

    item= gtk_menu_item_new_with_label("");
    gtk_widget_show(item);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);

    for (i= 0; i < connection_list->user_connections_num; i++)
    {
      conn= connection_list->user_connections+i;
      
      if (conn->storage_type == MYX_FAVORITE_USER_CONNECTION)
      {
        if (!conn->connection_name)
        {
          tmp= g_malloc(len+8);
          sprintf(tmp, _("* Unnamed %i *"), unnamed++);
          
          conn->connection_name= tmp;
        }

        item= gtk_menu_item_new_with_label(conn->connection_name);
        gtk_widget_show(item);
        gtk_object_set_data(GTK_OBJECT(item), "user_data", conn);

        gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);
      }

      if (conn->username)
          myg_entry_history_add(username_e, conn->username);
      if (conn->hostname)
          myg_entry_history_add(hostname_e, conn->hostname);
      if (conn->schema)
          myg_entry_history_add(schema_e, conn->schema);
    }
    dlg->connections = connection_list;
    
    /* set the last connection */
    if (connection_list->last_connection >= 0)
    {
      conn= connection_list->user_connections + connection_list->last_connection;
      wid= glade_xml_get_widget(dlg->xml, "connection_list");
      gtk_option_menu_set_history(GTK_OPTION_MENU(wid), 
                                  connection_list->last_connection);
    
      dlg->cur_conn= conn;
    }
    else
    {
      dlg->cur_conn= conn;
    }
  }
  else
  {
    g_warning("error loading connection list %s (%i)", clpath, merror);

    connection_list= g_malloc(sizeof(MYX_USER_CONNECTIONS));
    connection_list->user_connections_num= 0;
    connection_list->last_connection= -1;
    connection_list->user_connections= NULL;
    
    dlg->connections = connection_list;
  }
}


static void update_connect_sensitivity(myg_connect_dialog *dlg)
{
  int can_connect= 1;
  GtkWidget *wid;
  const char *text;
  
  /* check if everything is filled and we can connect */
  
  wid= glade_xml_get_widget(dlg->xml, "username_entry");
  text= gtk_entry_get_text(GTK_ENTRY(wid));
  if (!text || is_str_blank(text))
      can_connect= 0;
  
  wid= glade_xml_get_widget(dlg->xml, "hostname_entry");
  text= gtk_entry_get_text(GTK_ENTRY(wid));
  if (!text || is_str_blank(text))
      can_connect= 0;

  if (dlg->pick_schema)
  {
    wid= glade_xml_get_widget(dlg->xml, "schema_entry");
    text= gtk_entry_get_text(GTK_ENTRY(wid));
    if (!text || is_str_blank(text))
        can_connect= 0;
  }

  /* enable/disable button */
  wid= glade_xml_get_widget(dlg->xml, "connect_button");
  gtk_widget_set_sensitive(wid, can_connect);
}


static void *connection_thread(void *data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;
  
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  /* check if we got a cancelation before setting it async */
  pthread_testcancel();

  dlg->conn_result= -1;
  dlg->conn_result= myx_connect_to_instance(dlg->cur_conn, dlg->mysql);
  dlg->connecting= 0;
  
  return NULL;
}


static void process_ping_output(gpointer data, gint source, 
                                GdkInputCondition cond)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;
  char buffer[256];

  if (fgets(buffer, sizeof(buffer), dlg->ping_output))
  {
    myg_text_append(dlg->ping_view, buffer);                                 
  }
  else
  {
    /* error or end of data, simulate click on stop button */
    gtk_dialog_response(dlg->ping_dialog, 1);
  }
}


static void show_connection_failed(myg_connect_dialog *dlg)
{
  GtkWidget *mbox;
  GtkWidget *win;
  GtkDialog *dialog;
  GtkWidget *btn;
  pid_t ping_pid= 0;
  guint input_handler= 0;
  int rc;
  int pinging= 0;
  char title[strlen(dlg->app_title)+strlen(_("Error"))+4];
  char *ping_args[]= 
  {
    "/bin/ping",
        NULL,
        NULL
  };
  
  sprintf(title, "%s %s", dlg->app_title, _("Error"));

  win= glade_xml_get_widget(dlg->xml, "connect_dialog");

  mbox= gtk_message_dialog_new(GTK_WINDOW(win), 
                               GTK_DIALOG_MODAL
                               |GTK_DIALOG_DESTROY_WITH_PARENT,
                               GTK_MESSAGE_ERROR,
                               GTK_BUTTONS_CLOSE,
                               _("Could not connect to host '%s'.\n"
                                 "MySQL Error Nr. %i\n"
                                 "%s\n\n"
                                 "Click the 'Ping' button to see if there is a networking problem."),
                               dlg->cur_conn->hostname,
                               myx_mysql_errno(dlg->mysql),
                               myx_mysql_error(dlg->mysql));

  dialog= GTK_DIALOG(mbox);

  dlg->ping_dialog= dialog;
  
  btn= gtk_dialog_add_button(dialog, _("_Ping Host"), 1);

  while ((rc=gtk_dialog_run(dialog)))
  {
    if (pinging || rc == GTK_RESPONSE_CLOSE)
    {
      pinging= 0;
      gtk_button_set_label(GTK_BUTTON(btn), _("_Ping Host"));
      
      if (dlg->ping_output)
      {
        gtk_input_remove(input_handler);
        myx_pclose(dlg->ping_output, ping_pid);
        dlg->ping_output= NULL;
        ping_pid= 0;
      }
      
      if (rc == GTK_RESPONSE_CLOSE)
          break;
    }
    else
    {
      if (!dlg->ping_view)
      {
        dlg->ping_view= myg_add_text_to_dialog(dialog);
      }
      else
      {
        GtkTextBuffer *text;
        text= gtk_text_view_get_buffer(GTK_TEXT_VIEW(dlg->ping_view));
        gtk_text_buffer_set_text(text, "", -1);
      }
      
      if (dlg->ping_output)
      {
        myx_pclose(dlg->ping_output, ping_pid);
      }
      
      ping_args[1]= dlg->cur_conn->hostname;
      
      dlg->ping_output= myx_popen(ping_args, &ping_pid);
      if (!dlg->ping_output)
      {
        myg_text_append(dlg->ping_view, _("Could not execute ping\n"));
      }
      else
      {
        input_handler= gtk_input_add_full(fileno(dlg->ping_output),
                                          GDK_INPUT_READ,
                                          process_ping_output,
                                          NULL,
                                          (gpointer)dlg,
                                          NULL);
        
        pinging= 1;
        gtk_button_set_label(GTK_BUTTON(btn), _("Stop _Ping"));
      }
    }
  }

  gtk_widget_destroy(mbox);

  dlg->ping_dialog= NULL;
  dlg->ping_view= NULL;
}


static gint monitor_connection_thread(gpointer data)
{
  myg_connect_dialog *dlg= (myg_connect_dialog*)data;

  if (!dlg->connecting)
  {
    GtkWidget *wid;
    void *dummy;
    const char **p;

    /* finished connecting */

    pthread_join(dlg->conn_thread, &dummy);
    dlg->conn_thread= 0;

    /* reenable interface elements */
    for (p= disable_while_connecting; *p; p++) {
      wid= glade_xml_get_widget(dlg->xml, *p);
      gtk_widget_set_sensitive(wid, TRUE);
    }

    wid= glade_xml_get_widget(dlg->xml, "progress_image");
    gtk_widget_hide(wid);

    if (dlg->conn_result == 0)
    {
      /* success! */
      gtk_main_quit();
    }
    else
    {
      show_connection_failed(dlg);
      
      myx_mysql_close(dlg->mysql);
      dlg->mysql= NULL;
    }

    return 0;
  }

  return 1;
}


static void start_connection(myg_connect_dialog *dlg)
{
  GtkWidget *wid;
  const char **p;

  dlg->cur_conn= fetch_connection_values(dlg, dlg->cur_conn);

  /* disable interface elements while connecting... */
  for (p= disable_while_connecting; *p; p++) 
  {
    wid= glade_xml_get_widget(dlg->xml, *p);
    gtk_widget_set_sensitive(wid, FALSE);
  }
  
  wid= glade_xml_get_widget(dlg->xml, "progress_image");
  gtk_widget_show(wid);

  if (!dlg->conn_thread)
  {
    dlg->connecting= 1;

    dlg->mysql= myx_mysql_init();
    /* start connection thread */
    pthread_create(&dlg->conn_thread, NULL, connection_thread, dlg);

    gtk_idle_add(monitor_connection_thread, dlg);
  }
}


static void stop_connection(myg_connect_dialog *dlg)
{
  GtkWidget *wid;
  const char **p;
  
  /* stop connection thread */
  if (dlg->conn_thread)
  {
    void *ret;
    pthread_cancel(dlg->conn_thread);
    pthread_join(dlg->conn_thread, &ret);
    dlg->conn_thread= 0;
  }

  if (dlg->mysql)
  {
    myx_mysql_close(dlg->mysql);
    dlg->mysql= NULL;
  }
  
  /* reenable interface elements */
  for (p= disable_while_connecting; *p; p++) {
    wid= glade_xml_get_widget(dlg->xml, *p);
    gtk_widget_set_sensitive(wid, TRUE);
  }
  
  wid= glade_xml_get_widget(dlg->xml, "progress_image");
  gtk_widget_hide(wid);
}





static void reset_connection_values(myg_connect_dialog *dlg)
{
  GtkWidget *wid;

  wid= glade_xml_get_widget(dlg->xml, "connection_list");
  gtk_option_menu_set_history(GTK_OPTION_MENU(wid), 0);

  wid= glade_xml_get_widget(dlg->xml, "username_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), "");
  
  wid= glade_xml_get_widget(dlg->xml, "password_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), "");

  wid= glade_xml_get_widget(dlg->xml, "hostname_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), "");

  wid= glade_xml_get_widget(dlg->xml, "schema_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), "");

  wid= glade_xml_get_widget(dlg->xml, "port_spin");
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(wid), MYSQL_PORT);
}


static void display_connection_values(myg_connect_dialog *dlg,
                                      MYX_USER_CONNECTION *conn)
{
  GtkWidget *wid;
  
  wid= glade_xml_get_widget(dlg->xml, "username_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), (char*)conn->username?:"");
  
  wid= glade_xml_get_widget(dlg->xml, "password_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), (char*)conn->password?:"");
  if (!conn->password || !*conn->password)
  {
    gtk_widget_activate(wid);
  }

  wid= glade_xml_get_widget(dlg->xml, "hostname_entry");
  gtk_entry_set_text(GTK_ENTRY(wid), (char*)conn->hostname?:"");
  
  wid= glade_xml_get_widget(dlg->xml, "port_spin");
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(wid), conn->port);
  
  if (dlg->pick_schema)
  {
    wid= glade_xml_get_widget(dlg->xml, "schema_entry");
    gtk_entry_set_text(GTK_ENTRY(wid), (char*)conn->schema?:"");
  }
}



static MYX_USER_CONNECTION *fetch_connection_values(myg_connect_dialog *dlg,
                                                   MYX_USER_CONNECTION *conn)
{
  GtkWidget *wid;

  if (!conn)
  {
    conn= g_malloc(sizeof(MYX_USER_CONNECTIONS));
    memset(conn, 0, sizeof(MYX_USER_CONNECTIONS));
    
    conn->storage_type= MYX_HISTORY_USER_CONNECTION;
  }

  wid= glade_xml_get_widget(dlg->xml, "username_entry");
  if (conn->username) free(conn->username);
  conn->username= ZSTRDUP(gtk_entry_get_text(GTK_ENTRY(wid)));

  wid= glade_xml_get_widget(dlg->xml, "password_entry");
  if (conn->password) free(conn->password);
  conn->password= ZSTRDUP(gtk_entry_get_text(GTK_ENTRY(wid)));

  wid= glade_xml_get_widget(dlg->xml, "hostname_entry");
  if (conn->hostname) free(conn->hostname);
  conn->hostname= ZSTRDUP(gtk_entry_get_text(GTK_ENTRY(wid)));

  wid= glade_xml_get_widget(dlg->xml, "port_spin");
  conn->port= gtk_spin_button_get_value(GTK_SPIN_BUTTON(wid));

  wid= glade_xml_get_widget(dlg->xml, "schema_entry");
  if (conn->schema) free(conn->schema);
  conn->schema= ZSTRDUP(gtk_entry_get_text(GTK_ENTRY(wid)));
  
  return conn;
}




#ifdef TEST
int main(int argc, char **argv)
{
  gtk_init(&argc, &argv);
  glade_init();
  
  myg_connect_with_dialog("../../res/linux/connect_dialog.glade",
                          "/home/alfredo/.mysqlgui/mysqlx_user_connections.xml",
                          "MySQL Foobar Tool", 0);

  return 0;
}
#endif

