unit InstanceDataModule;

interface

uses
  Windows, Messages, SysUtils, Classes, ComCtrls, ImgList, Controls, Forms, Contnrs,
  ExtCtrls, AuxFuncs, PNGImage, StdCtrls, SyncObjs, ApplicationDataModule,
  myx_public_interface, myx_admin_public_interface, MyxError;

const
  WM_Disconnected = WM_USER+101;
  WM_Reconnected = WM_USER+102;

type
  TFetchDataThread = Class;

  DataBeingFetchedSet = set of 1..100;

  TThreadExecMethod = procedure(Sender: TObject) of object;

  TInstanceDM = class(TDataModule)
    constructor Create(AOwner: TComponent; StatusBar: TStatusbar); reintroduce;

    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);

    function ConnectToServer(MySQLConnectionsOnly: Boolean = True;
      SelectSchemata: Boolean = False): integer;

    procedure Reconnect;
    procedure Disconnect;

    procedure FetchData(KindOfData: integer;
      FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
      Target: TObject; StatusBarText: string;
      AllowParallelExecution: Boolean = False);
    procedure TerminateFetchingData;
  private
    { Private declarations }

  public
    { Public declarations }
    //Pointer to MySQL Struct
    PMySQL: Pointer;
    MySQLMajorVersion: integer;

    //Pointer to MYX_USER_CONNECTION
    user_connection: TMYX_USER_CONNECTION;

    ConnectionCaption: string;

    //Fetch Data Thread
    FetchDataThreadHandle: THandle;
    FetchDataThread: TFetchDataThread;

    FetchDataThreadCriticalSection: TCriticalSection;
    FetchDataMultiReadSync: TMultiReadExclusiveWriteSynchronizer;

    DataBeingFetched: DataBeingFetchedSet;
    FetchingData: Boolean;

    StatusBar: TStatusbar;

    Connected: Boolean;
    ConnectedToLocalhost: Boolean;
  end;

  TFetchDataThread = class(TThread)
    constructor Create(InstanceDM: TInstanceDM; KindOfData: integer;
      FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
      Target: TObject; StatusBarText: string; AllowParallelExecution: Boolean);
    destructor Destroy; override;
    procedure ReleaseCriticalSection;
    procedure UpdateStatusBar;
    procedure ShowError;
    procedure ExecuteSynchronized(SynchronizedMethod: TThreadExecMethod);
    procedure ExecuteSynchronizedMethod;
  protected
    procedure Execute; override;
  private
    ErrorMsg: string;
    FetchMethod: TThreadExecMethod;
    RefreshMethod: TThreadExecMethod;
    SynchronizedMethod: TThreadExecMethod;
  public
    InstanceDM: TInstanceDM;
    KindOfData: integer;
    Target: TObject;
    StatusBarText: string;
    AllowParallelExecution: Boolean;
  end;

const
  //KindOfData
  KindOfData_CatalogSchema = 1;
  KindOfData_UserNames = 2;
  KindOfData_ProcessList = 3;
  KindOfData_SchemaTables = 4;
  KindOfData_StatusVars = 5;
  KindOfData_ServerVars = 6;
  KindOfData_UserData = 7;
  KindOfData_SchemaTableStatus = 8;
  KindOfData_TableAction = 9;

implementation

uses Main, ConnectToInstance;

{$R *.dfm}

constructor TInstanceDM.Create(AOwner: TComponent; StatusBar: TStatusbar);
begin
  inherited Create(AOwner);

  self.StatusBar:=StatusBar;
end;

procedure TInstanceDM.DataModuleCreate(Sender: TObject);
begin
  //DLL Version check

  PMySQL:=nil;
  user_connection:=nil;

  Connected:=False;
  ConnectedToLocalhost:=False;

  //Initialize components for thread handling
  FetchDataThreadCriticalSection:=TCriticalSection.Create;
  FetchDataMultiReadSync:=TMultiReadExclusiveWriteSynchronizer.Create;
  DataBeingFetched:=[];

  FetchDataThread:=nil;
  FetchingData:=False;
end;

procedure TInstanceDM.DataModuleDestroy(Sender: TObject);
begin
  if(FetchingData)then
    TerminateThread(FetchDataThreadHandle, 0);

  if(FetchDataThread<>nil)then
    FetchDataThread.Free;

  FetchDataMultiReadSync.Free;
  FetchDataThreadCriticalSection.Free;

  //Close connection and free memory for MySQL struct
  myx_mysql_close(PMySQL);
end;

function TInstanceDM.ConnectToServer(MySQLConnectionsOnly: Boolean = True;
  SelectSchemata: Boolean = False): integer;
var ConnectToInstanceForm: TConnectToInstanceForm;
  ConnnectionResult: integer;
begin
  ConnectToInstanceForm:=TConnectToInstanceForm.Create(self,
    MySQLConnectionsOnly, SelectSchemata);
  try
    ConnnectionResult:=ConnectToInstanceForm.ShowModal;
    if(ConnnectionResult=mrOK)then
    begin
      //Initialize Variables
      User_connection:=ConnectToInstanceForm.User_Connection;
      PMySQL:=ConnectToInstanceForm.PMySQL;
      MySQLMajorVersion:=myx_get_mysql_major_version(ConnectToInstanceForm.PMySQL);

      ConnectionCaption:=
        ConnectToInstanceForm.User_Connection.username+'@'+
        ConnectToInstanceForm.User_Connection.hostname+':'+
        IntToStr(ConnectToInstanceForm.User_Connection.port);

      ConnectToServer:=1;

      Connected:=True;
      ConnectedToLocalhost:=
        (myx_is_localhost(PChar(AnsiToUtf8(user_connection.hostname)))=1);
    end
    else if(ConnnectionResult=mrAbort)then
      ConnectToServer:=-1
    else
      ConnectToServer:=0;

  finally
    ConnectToInstanceForm.Free;
  end;
end;

procedure TInstanceDM.FetchData(KindOfData: integer;
  FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
  Target: TObject; StatusBarText: string; AllowParallelExecution: Boolean);
begin
  if(Not(KindOfData in DataBeingFetched))then
  begin
    //Add KindOfData to DataBeingFetched Set
    DataBeingFetched:=DataBeingFetched + [KindOfData];
    try
      //Create a new thread
      FetchDataThread:=TFetchDataThread.Create(self,
        KindOfData, FetchMethod, RefreshMethod,
        Target, StatusBarText, AllowParallelExecution);
      try
        //Thread will be freed after execution
        FetchDataThread.FreeOnTerminate:=True;

        FetchingData:=True;

        FetchDataThread.Resume;
      except
        FetchDataThread.Free;
        FetchDataThread:=nil;

        raise;
      end;
    except
      DataBeingFetched:=DataBeingFetched - [KindOfData];

      raise;
    end;
  end;
end;

procedure TInstanceDM.TerminateFetchingData;
begin
  TerminateThread(FetchDataThreadHandle, 0);
  FetchDataThreadCriticalSection.Release;
  DataBeingFetched:=DataBeingFetched - [FetchDataThread.KindOfData];

  if(FetchDataThread<>nil)then
    FetchDataThread.Free;
  FetchDataThread:=nil;
  
  FetchingData:=False;

  if(StatusBar<>nil)then
  begin
    StatusBar.Panels[1].Text:='';
    StatusBar.Invalidate;
  end;
end;

procedure TInstanceDM.Reconnect;
var ConnectionResult: integer;
begin
  if(User_Connection=nil)then
    Exit;

  if(PMySQL<>nil)then
    Disconnect;

  PMySQL:=myx_mysql_init();
  if(PMySQL=nil)then
    raise EMyxError.Create('Error while allocating memory for MySQL Struct.');

  ConnectionResult:=myx_connect_to_instance(
    User_Connection.get_record_pointer,
    PMySQL);
  if(ConnectionResult<>0)then
    raise EMyxSQLError.Create('Could not connect to MySQL Server',
      myx_mysql_errno(PMySQL), myx_mysql_error(PMySQL));

  Connected:=True;
  MessageToAllForms(WM_Reconnected, 0, 0);
end;

procedure TInstanceDM.Disconnect;
begin
  //Close connection and free memory for MySQL struct
  myx_mysql_close(PMySQL);
  PMySQL:=nil;

  Connected:=False;
  MessageToAllForms(WM_Disconnected, 0, 0);
end;



//------------------------------------------------------------------------------

constructor TFetchDataThread.Create(
  InstanceDM: TInstanceDM; KindOfData: integer;
  FetchMethod: TThreadExecMethod; RefreshMethod: TThreadExecMethod;
  Target: TObject; StatusBarText: string; AllowParallelExecution: Boolean);
begin
  inherited Create(True);

  self.InstanceDM:=InstanceDM;
  self.KindOfData:=KindOfData;
  self.FetchMethod:=FetchMethod;
  self.RefreshMethod:=RefreshMethod;
  self.Target:=Target;
  self.StatusBarText:=StatusBarText;
  self.SynchronizedMethod:=nil;
  self.AllowParallelExecution:=AllowParallelExecution;
end;

destructor TFetchDataThread.Destroy;
begin
  InstanceDM.FetchingData:=False;
  StatusBarText:='';
  Synchronize(UpdateStatusBar);

  InstanceDM.FetchDataThread:=nil;

  InstanceDM.DataBeingFetched:=InstanceDM.DataBeingFetched - [KindOfData];

  inherited Destroy;
end;

procedure TFetchDataThread.Execute;
begin
  //Enter CriticalSection
  if(Not(AllowParallelExecution))then
    InstanceDM.FetchDataThreadCriticalSection.Acquire;
  try
    try
      InstanceDM.FetchDataThreadHandle:=Handle;

      if(StatusBarText<>'')then
        Synchronize(UpdateStatusBar);

      //Test long response time
      //sleep(3000);

      if(Assigned(FetchMethod))then
        FetchMethod(self);

      {if(StatusBarText<>'')then
      begin
        StatusBarText:='';
        InstanceDM.FetchingData:=False;
        Synchronize(UpdateStatusBar);
      end;}

      //The Refresh Method is called Synchronized
      if(Assigned(RefreshMethod))then
        ExecuteSynchronized(RefreshMethod);
    except
      on x: EMyxSQLError do
      begin
        ErrorMsg:=x.Message+#13#10#13#10+
          'MySQL Error Nr.'+EMyxSQLError(x).ErrorNr+#13#10+
          EMyxSQLError(x).ErrorTxt;
        Synchronize(ShowError);
      end;
      on x: Exception do
      begin
        ErrorMsg:=x.Message;
        Synchronize(ShowError);
      end;
    end;
  finally
    if(Not(AllowParallelExecution))then
      InstanceDM.FetchDataThreadCriticalSection.Release;
  end;
end;

procedure TFetchDataThread.ReleaseCriticalSection;
begin
  InstanceDM.FetchDataThreadCriticalSection.Release;
end;

procedure TFetchDataThread.UpdateStatusBar;
begin
  if(InstanceDM.StatusBar<>nil)then
  begin
    InstanceDM.StatusBar.Panels[1].Text:=StatusBarText;
    InstanceDM.StatusBar.Repaint;
  end;
end;

procedure TFetchDataThread.ShowError;
begin
{$IFNDEF PresentationMode}
  ShowModalDialog(Application.Title+' Error', ErrorMsg,
    myx_mtError, 'OK');
{$ENDIF}
end;

procedure TFetchDataThread.ExecuteSynchronized(SynchronizedMethod: TThreadExecMethod);
begin
  self.SynchronizedMethod:=SynchronizedMethod;

  Synchronize(ExecuteSynchronizedMethod);
end;

procedure TFetchDataThread.ExecuteSynchronizedMethod;
begin
  if(@SynchronizedMethod<>nil)then
    SynchronizedMethod(self);
end;


end.
